"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes.oldleedpattern ***

Defines the LEEDPattern and LEEDsubpattern classes, used for displaying a
LEEDPattern. This is the old version of the classes. Reimplemented with new
logics in leedpattern.py. This module is kept until the reimplementation is
ready.

Author: Michele Riva
Created: 2021-03-13
"""

import numpy as np
from quicktions import Fraction
from scipy import spatial as sp
from matplotlib import cm  # color maps
from matplotlib import colors as mpl_colors

from viperleed import guilib as gl


class LEEDPattern:
    # @gl.profile_calls(print_args=[30])
    def __init__(self, leed_parameters):
        # run leed_parameters through gl.LEEDParameters, so that all
        # parameters are correctly interpreted, and missing optional values
        # are set to their default
        self.__parameters = gl.LEEDParameters(leed_parameters)

        # The following attributes are set by __build_lattices() and
        # __build_LEEDPattern()
        # * reciprocal-space lattices for surface and bulk
        self.reciprocal_lattices = {'surf': None, 'bulk': None}

        # * characteristics of domains generated by symmetry operations
        #   of the bulk. 'g' is a list (one element per domain) containing
        #   all the reciprocal-space wave-vectors (in x,y Cartesian coordinates)
        #   of the beams from one domain
        self.domains = {'operations': None, 'g': None}

        ## NEED TO REDUCE THE NUMBER OF ATTRIBUTES FROM HERE ON
        self.surfLEED = None  # like self.domains['g'], but flattened
        self.superposedBeams = None  # grouping of beam indices telling which ones overlap   ## PROBABLY WILL GO with BeamIndexList
        self.names = None     # strings of beam names (wrt bulk basis)            ## REPLACE via BeamIndexList!

        # The following one is related to symmetry-equivalent beams
        self.firstDomRefs = None  # references of equivalent beams in 1st domain  ## REPLACE (different logics for off normal).
        
        # The following ones are related to the LEED subpatterns                  ## THE LOGICS NEEDS TO CHANGE!
        # Will probably have a self.patterns dictionary that is initially populated only with some of the subpatterns. Other combinations will be dynamically generated and stored as new keys, with some reasonable hashing.
        self.firstLEED = None  # subpattern of first domain
        self.domsLEED = None   # subpatterns for all domains
        self.domColors = None  # colors for plotting multiple domains

        self.__build_lattices()
        self.__build_LEEDPattern()

    @property
    def bulk_basis(self):
        if self.reciprocal_lattices['bulk'] is None:
            return np.dot(self.superlattices[0].T,                              # ENOUGH TO HAVE THE FIRST ONE
                          self.reciprocal_lattices['surf'].basis)
        return self.reciprocal_lattices['bulk'].basis

    @property
    def max_energy(self):
        """
        Maximum LEED energy
        """
        return self.__parameters['eMax']

    @property
    def n_beams(self):                                                          # NOT CORRECT FOR MULTIPLE STRUCTURES. Probably not needed
        """
        Number of LEED beams in the pattern
        """
        return self.reciprocal_lattices['surf'].n_beams

    @property
    def n_domains(self):
        """
        Number of distinct domains produced by the symmetry operations of the
        bulk
        """
        return len(self.domains['operations'])

    @property
    def superlattices(self):
        """
        Numpy ndarray of superlattice matrices generating the symmetry-related
        domains. The first element is the one generating the lattice whose basis
        is given in the constructor as a LEED parameter
        """
        superlattice = self.__parameters['SUPERLATTICE']
        operations = self.domains['operations']
        if operations is None:
            operations = [gl.PlaneGroup.E]  # identity matrix
        # return np.asarray([np.dot(superlattice, operation)
                           # for operation in operations])
        return np.einsum('ij,mjk->mik', superlattice, operations)
    
    @property
    def primary_beam_theta(self):
        """
        Polar angle of incidence with respect to the direction perpendicular
        to the surface
        """
        return self.__parameters['beamIncidence'][0]
    
    @property
    def primary_beam_phi(self):
        """
        Azimuthal angle of incidence of the primary beam with respect to the
        x axis. Positive counterclockwise when looking down on the surface.
        """
        return self.__parameters['beamIncidence'][1]
    
    @property
    def primary_beam_direction(self):                                           # will probably not even use this
        """
        2-tuple representing the in-plane direction vector of incidence,
        in Cartesian coordinates
        """
        if self.primary_beam_theta == 0:
            return (0, 0)
        phi = np.radians(self.primary_beam_phi)
        return np.cos(phi), np.sin(phi)

    def __build_lattices(self):                                                 # Not even sure if we really need both lattices anymore. Probably bulk + domains are enough
        # Create dummy surface lattice just to get the reciprocal basis
        surf = gl.Lattice(self.__parameters['surfBasis'])

        # and get the maximum screen radius
        max_radius = self.screen_radius(self.max_energy)

        # Now make the reciprocal lattices:
        self.reciprocal_lattices['surf'] = gl.Lattice(
            surf.reciprocal_basis, space='reciprocal',
            group=self.__parameters['surfGroup'], limit=max_radius
            )
        self.reciprocal_lattices['bulk'] = gl.Lattice(
            self.bulk_basis, space='reciprocal',
            group=self.__parameters['bulkGroup'], limit=max_radius
            )
        self.reciprocal_lattices['bulk'].group.screws_glides = (
            self.__parameters['bulk3Dsym'],
            self.reciprocal_lattices['bulk'].cell_shape
            )

    # @gl.profile_lines
    def __build_LEEDPattern(self):                                                ## bottleneck!

        # 1) get the (bulk) operations that generate symmetry-equivalent domains
        self.domains['operations'] = self.__find_domain_operations()

        # 2) generate the gx gy coordinates for all the domains                   ## PERHAPS IT MAKES SENSE TO GENERATE hk(s) FIRST? THESE CAN BE CONVERTED TO a BeamIndexList
        self.domains['g'] = [                                                     ## NEEDS TO CHANGE TO ACCOUNT FOR OFF-NORMAL. It may also be that I don't need this at all,
            np.linalg.multi_dot(
                (self.reciprocal_lattices['surf'].hk,
                 np.linalg.inv(superlattice).T,
                 self.bulk_basis)
                )
            for superlattice in self.superlattices                                ## CAN BE PART OF LEEDDomains
            ]

        # 3) prepare the full LEED pattern of the surface, including all
        # domains
        surfLEEDx = np.array([self.domains['g'][i][:, 0]
                              for i in range(self.n_domains)]).flatten()
        surfLEEDy = np.array([self.domains['g'][i][:, 1]
                              for i in range(self.n_domains)]).flatten()
        self.surfLEED = np.array(list(zip(surfLEEDx, surfLEEDy)))

        # 4) find superposed beams
        self.superposedBeams = self.get_supSpots()

        # 5) prepare the formatted names of the fractional indices
        self.names = np.array([
            self.formatFractionalIndices(self.reciprocal_lattices['surf'].hk, m)  # MAIN BOTTLENECK: > half time of whole method
            for m in self.superlattices])                                         # WILL NOT EXIST

        # 6) and build the subpatterns
        self.__build_subpatterns()                                                ## BOTTLENECK: half of time of whole method

    def get_supSpots(self):                                                       ## TO EVALUATE WHETHER IT'S FASTER TO USE BeamIndexList
        # Group all the beams into separate groups, such that each group
        # contains the indices of overlapping beams (can be one index if non
        # overlapping)
        tree = sp.cKDTree(self.surfLEED)
        return np.array(tree.query_ball_point(x=self.surfLEED, r=1e-8))

    # @gl.profile_calls(print_args=[30])
    def get_beamGrouping(self, domains=None):                                     ## TO RETHINK FOR NON-NORMAL INCIDENCE  ## BOTTLENECK for __build_subpatterns
        # this is wrong for non-normal incidence!
        # all domains should be treated separately, rather than using the
        # symmetry of the first one, since the special direction given by the
        # beam incidence has a different effect on each domain
        self.firstDomRefs = self.get_FirstDomainSymmetry()                        ## BOTTLENECK?

        if domains is None:
            domains = range(self.n_domains)

        # beamsRefs contains at index [i, j] a list of the indices of beams of
        # domain domains[i] equivalent to beam j (including j)
        beamsRefs = []
        for dom in domains: 
            beamsRefs.append([crossRef + dom*self.n_beams
                              for crossRef in self.firstDomRefs])
        return np.array(beamsRefs, dtype=object)

    def get_equivalentSpots(self, domains=None):                                  ## REPLACE with generic function!
        if not hasattr(domains, '__len__') and domains is not None:
            raise TypeError("Invalid type for domains. "
                            "Should be None or an array-like.")
        if domains is None:
            # output spots from all domains
            domains = range(self.n_domains)
        elif not all(isinstance(dom, int) for dom in domains):
            raise ValueError("Invalid domain index. "
                             "All indices should be integers")
        elif any(dom < 0 or dom >= self.n_domains for dom in domains):
            raise ValueError("Domain index out of range. "
                             "Indices should be between 0 "
                             f"and {self.n_domains}.")

        names = self.names[domains]
        beamRefs = self.get_beamGrouping(domains)

        # get ready to prepare one dictionary per each domain:
        # key: beam name
        # value: names of beams equivalent to key
        #        Notice that for each key (=beam) in one domain, one should
        #        also include as equivalent beams those from other domains
        #        at the same positional index.
        #        !!!THIS IS A PROBLEM FOR NON-NORMAL INCIDENCE!!!
        #        --> get this list (the same for all domains) ready first
        allRefs = [np.concatenate(beamRefs[:, beam]).astype(int)
                   for beam in range(self.n_beams)]
        refNames = [set(self.names.ravel()[beamRef]) for beamRef in allRefs]

        # now pack the dictionaries
        # Notice that it is necessary to have one dictionary per domain since
        #   some domains will give rise to a subset of all the LEED spots:
        #   e.g., in a (2x1) on square bulk, one domain will give spot (1/2, 0)
        #   the other domain spot (0, 1/2)
        # Having all the dictionaries allows to determine which of the
        #   superposed spots are indeed equivalent (see next comment) as well
        #   as which domains overlap
        domsDicts = [dict(zip(domNames, refNames)) for domNames in names]

        # now figure out which of the beams are symmetry equivalent
        # taking superposition of spots into account:
        # spots originating from beams that are symmetry equivalent within
        # each domain, are symmetry equivalent only if all domains contribute
        # with symmetry-equivalent beams in all spots.
        #
        # For example (hexagonal bulk):
        # dom1: (1 0), (-1 0), (1 -1), (-1 1) equivalent
        # dom2: (1 0), (-1 0), (0 -1), (0 1)  equivalent
        # dom3: (1 0), (-1 0)                 equivalent
        # --> (1 0) and (-1 0) are equivalent
        #
        # dom1: (1 0),  (-1 0),  (1 -1), (-1 1)    equivalent
        # dom2:                  (1 -1), (-1 1)    equivalent
        # dom3: (0, 1), (0, -1), (1 -1), (-1 1)    equivalent
        # --> (1 -1) and (-1 1) are equivalent, but are
        #     not equivalent to [(1 0), (-1 0)]
        eqBeams = []
        for name in names.ravel():
            setLst = [dDict[name] for dDict in domsDicts if name in dDict]
            if setLst:
                newBeams = set.intersection(*setLst)
                # remove the keys of the beams that have
                # already been processed from each domain
                for domDict in domsDicts:
                    for beam in newBeams:
                        if beam in domDict:
                            del domDict[beam]
                eqBeams.append(newBeams)

        # sort within each equivalence group
        lstBeams = [sorted(list(beams), key=self.beamsSortCriterion)
                    for beams in eqBeams]
        # and by energy
        sortedBeams = sorted(lstBeams, key=self.sortEnergy)

        beamsWithIndices = []
        overlapDoms = []
        for (i, beams) in enumerate(sortedBeams):
            for beam in beams:
                overlapDoms = [dom + 1 for dom in domains
                               if beam in self.names[dom].ravel()]
                beamsWithIndices.append([beam, i, overlapDoms])

        # Now find which of the beams are extinct due to glide, place them
        # in a list of beam indices in the same order as in self.names[i]
        extinct = self.get_ExtinctFirstDomain()

        extinctNames = self.names[:, extinct]
        # notice that extinctNames is an empty list if there are no glide
        # extinct beams.
        for b, (beam, group, doms) in enumerate(beamsWithIndices):
            extDoms = [dom for dom in doms if beam in extinctNames[dom-1]]
            # Notice the dom-1 because of the way the doms list is
            # built in the previous for loop

            beamsWithIndices[b] = [beam, group, doms, extDoms]
            # now, only in case extDoms is of length 1, i.e., one domain
            # only gives rise to the extinct beam, the group index goes
            # negative
            # !!! Perhaps will be changed with the better version in base.py !!!
            if len(extDoms) == 1:
                beamsWithIndices[b][1] *= -1

        return beamsWithIndices

    def beamsSortCriterion(self, beamstr):                                        ## this may go into BeamIndexList
        beam = eval(beamstr)  # returns a tuple                                   ## This will go

        sortH = beam[0]
        sortK = beam[1]

        return (-sortH-sortK, -sortH)
    
    def sortEnergy(self, beamlist):                                               ## this may go into BeamIndexList
        beam = eval(beamlist[0])  # returns a tuple
        # sort by Energy, i.e., by g^2. Notice the
        # rounding to reduce floating-point errors
        g_norm = np.linalg.norm(np.dot(beam, self.bulk_basis)).round(7)
        return (g_norm, *self.beamsSortCriterion(beamlist[0]))

    # @gl.profile_lines
    def __build_subpatterns(self):
        # Now sort the sub-patterns:
        # - There will always be a firstLEED and a domsLEED
        #   This is to handle the plotting of the first domain only.              ## NEEDS TO CHANGE TO DISPLAY EACH DOMAIN SEPARATELY OR AN ARBITRARY COMBINATION OF DOMAINS
        #   All entries in these lists are of type LEEDsubpattern
        #
        # - firstLEED, contains at most two entries: the first one
        #   includes all non-extinct beams, the second one
        #   the extinct ones, if any are present
        # - domsLEED is empty if only one domain is present.
        #   Otherwise, it contains:
        #   * one entry with superimposed spots
        #   * at most two entries for each domain.
        #     first entry: non-extinct; second entry: extinct (if any)

        doms = self.domains['g']
        symEq = self.get_beamGrouping()                                          # BOTTLENECK
        names = self.names

        maskFirst = np.full(self.n_beams, True)
        maskFirst[self.get_ExtinctFirstDomain()] = False

        self.firstLEED = [LEEDsubpattern(doms[0][maskFirst],
                                         symEq[0][maskFirst],
                                         names[0][maskFirst])]
        if all(maskFirst):  # no glide-extinct beams
            pass
        else:
            glideAlpha = 0.2
            glideScale = 2
            maskExtinct = np.invert(maskFirst)
            self.firstLEED.append(LEEDsubpattern(doms[0][maskExtinct],
                                                 symEq[0][maskExtinct],
                                                 names[0][maskExtinct],
                                                 color='k',
                                                 alpha=glideAlpha,
                                                 marker='x',
                                                 sizeScale=glideScale
                                                 ))

        self.domsLEED = []
        self.domColors = None
        if self.n_domains > 1:
            # If there are domains, there always will be overlapping spots,
            # at least those at integer orders

            # superposedMask is True at all the beam indices of beams that
            # are superposed to others
            superposedMask = np.array([len(sup) > 1
                                       for sup in self.superposedBeams])

            # figure out which domains superpose:
            overlapDoms = [[x//self.n_beams + 1 for x in sup]
                           for sup in self.superposedBeams[superposedMask]]
            self.domsLEED = [LEEDsubpattern(self.surfLEED[superposedMask],
                                            symEq.ravel()[superposedMask],
                                            names.ravel()[superposedMask],
                                            domain=-1,
                                            color='gray',
                                            overlapDoms=overlapDoms
                                            )]

            # define colors
            self.domColors = cm.gnuplot(np.linspace(0.1, 0.9, self.n_domains))

            for dom in range(self.n_domains):
                mask = np.invert(
                    superposedMask[dom*self.n_beams:(dom+1)*self.n_beams])
                maskDom = np.logical_and(maskFirst, mask)
                self.domsLEED.append(LEEDsubpattern(doms[dom][maskDom],
                                                    symEq[dom][maskDom],
                                                    names[dom][maskDom],
                                                    domain=dom+1,
                                                    color=self.domColors[dom]
                                                    )
                                     )
                if all(maskFirst):  # no glide extinct
                    pass
                else:
                    maskDomExtinct = np.logical_and(maskExtinct, mask)
                    self.domsLEED.append(
                        LEEDsubpattern(doms[dom][maskDomExtinct],
                                       symEq[dom][maskDomExtinct],
                                       names[dom][maskDomExtinct],
                                       domain=dom+1,
                                       color=self.domColors[dom],
                                       alpha=glideAlpha,
                                       marker='x',
                                       sizeScale=glideScale))

    def screen_radius(self, en):
        return gl.screen_radius(en, self.__parameters['screenAperture'])

    def get_FirstDomainSymmetry(self):
        """
        Returns
        -------
        crossRefs: list of np.arrays,
                   length == len(self.reciprocal_lattices['surf'].hk)
                   Each np.array contains the positional indices of the
                   beams equivalent to the current one (including
                   self-reference)
        """

        hk = self.reciprocal_lattices['surf'].hk
        N = len(hk)
        symOps = self.reciprocal_lattices['surf'].group.operations()

        # at first, use sets to avoid duplicates
        crossRefs = [{i} for i in range(N)]

        for op in symOps:
            hk_transformed = np.dot(op, hk.T).T.tolist()
            # transform hk according to the operation of the group
            for i in range(N):
                idx = hk_transformed.index(hk[i].tolist())
                # find which transformed hk is equal to the current beam
                # NB: there's always only one for each (beam, operation) pair
                #
                # and add it to the set of cross references for the current
                # beam
                crossRefs[i] |= {idx}

        # then convert each set to a np.array of int, so that it's iterable
        # and it's easy to process later
        return np.array([np.array(list(x), dtype=int) for x in crossRefs], 
                        dtype=object)

    def get_ExtinctFirstDomain(self):
        """
        Returns
        -------
        list of indices of beams of the first domain that are extinct
        due to glide
        """
        hk = self.reciprocal_lattices['surf'].hk
        group = self.reciprocal_lattices['surf'].group.group

        if 'g' in group:  # group has glide
            if '[1 0]' in group:  # pg or pmg
                idx = np.where([x[1] == 0 and abs(x[0] % 2) == 1
                                for x in hk])[0].tolist()
            elif '[0 1]' in group:  # pg or pmg
                idx = np.where([x[0] == 0 and abs(x[1] % 2) == 1
                                for x in hk])[0].tolist()
            else:  # pgg or p4g
                idx = np.where([(x[1] == 0 and abs(x[0] % 2) == 1)
                                or (x[0] == 0 and abs(x[1] % 2) == 1)
                                for x in hk])[0].tolist()
        else:
            idx = []

        return idx

    def __find_domain_operations(self):
        """
        Finds the symmetry operations of the bulk lattice that give distinct
        domains in LEED.
        NB: this function gives different results than what one would get from
        LEEDPat, as we're interested also in the symmetry relations between the
        intensities of the spots, while LEEDPat cares only about the presence
        or not of any spot.

        Returns
        -------
        list of operations

        -------
        """
        # The current version is based on the concept of co-sets of a group.
        # Given a group G and a subgroup H, the left co-set of H with respect
        # to the group operation g of G is
        #           gH = {g*h : h in H}.
        # Additionally, we take into account that each element of G is found in
        # exactly only one co-set (e.g., H is the identity co-set).
        #
        # The operations g_i that generate distinct co-sets are those that will
        # generate distinct lattices

        def __array_2_tuple(arr):
            """
            Convenience function that is used to convert a 2x2 array to a 2x2
            tuple
            """
            return tuple(map(tuple, arr))

        # 1) project the operations from the surface group to the bulk; also
        #    rounding to integers
        transform = np.linalg.inv(self.superlattices[0])                        # FIRST ONE IS ENOUGH
        surf_ops = tuple(
            op.round().astype(int)
            for op
            in self.reciprocal_lattices['surf'].group.transform(transform)
            )

        # 2) keep track of the operations of the bulk group that are already in
        #    a co-set, and of the group operations giving distinct co-sets
        coset_ops = set()
        _ops = []

        # 3) run through the bulk operations, adding the operations of new
        #    co-sets to the set above
        bulk_group = self.reciprocal_lattices['bulk'].group
        bulk_ops = bulk_group.operations(include_3d=True)

        for b_op in bulk_ops:
            if b_op in coset_ops:
                continue
            _ops.append(b_op)
            coset = set(__array_2_tuple(np.dot(b_op, surf_op))
                        for surf_op in surf_ops)
            coset_ops.update(coset)

        return _ops

    def formatFractionalIndices(self, hk, m):                                     ## this will probably go, and will rather use BeamIndex; BOTTLENECK
        """
        Method that generates fractional indices of a superstructure m whose
        integer indices are hk.

        Parameters
        ----------
        hk, 2D-array
            Integer indices along the two directions to be formatted

        m, 2x2 np.array
            superlattice matrix

        Returns
        -------
        names, 1D array
            each element is a string with formatted names for the indices
        """

        # The beam indices [hb,kb] for the surface indices [h,k] are
        #    [hb, kb] = np.dot([h, k], G),
        # with G = np.linalg.inv(M).transpose()
        hkbulk = np.dot(np.linalg.inv(m), hk.T).T

        # now get the formatting done:
        # all fractional indices are of type nn/mu with nn integer, and
        mu = abs(np.round(np.linalg.det(m)).astype(int)).item()

        # notice that, to prevent Fraction().limit_denominator() from
        # overflowing I'm casting the numpy-type numbers to their native
        # python type via .item()
        names = [', '.join([str(Fraction(hh.item()).limit_denominator(mu))         ## BOTTLENECK is limit_denominator
                            for hh in hkb]) for hkb in hkbulk]

        return np.array(names)

    def rotate(self, angle, which='surf'):
        """
        This function is a mess and needs to be split into two functions.
        It currently does two different things: if which='bulk', it returns a
        copy of the bulk lattice. If which='surf' it rotates all the LEED
        subpatterns. It would probably be nicer to also have the bulk pattern
        be a special subpattern.
        """
        which = which[0:4]
        if which not in ['surf', 'bulk']:
            raise ValueError("LEEDPattern: Only 'surf', 'surface', or "
                             "'bulk' are acceptable values for rotate()")
        if hasattr(angle, '__len__'):
            raise ValueError('Angle input is not a scalar')

        if which == 'surf':
            angle = np.radians(angle)
            rot = np.array([[np.cos(angle), np.sin(angle)],
                            [-np.sin(angle), np.cos(angle)]])
            for pat in [*self.firstLEED, *self.domsLEED]:
                pat.transformCoords(rot)
            return None
        return self.reciprocal_lattices['bulk'].get_rotated_lattice(angle)


class LEEDsubpattern:
    """
    --> I originally intended to use this guy for the
         determination of beam equivalence and
         handling of hovering annotations
    """
    def __init__(self, beams, symBeams, names, domain=1, color='k',
                 alpha=None, marker='o', sizeScale=1, overlapDoms=None):
        if not all([len(beams) == len(x) for x in [symBeams, names]]):
            raise ValueError('Input shape mismatch. The length of symBeams'
                             'and/or _names differs from the one of beams.')
        if not isinstance(domain, int):
            raise ValueError("domain must be an integer.")

        self.beams = beams
        self.rotBeams = self.beams
        self.symmetricBeams = symBeams
        self.names = names
        self.domain = domain
        if alpha is None:
            self.color = np.array([mpl_colors.to_rgba(color, alpha=1)])
        else:
            self.color = np.array([mpl_colors.to_rgba(color, alpha=alpha)])
        self.marker = marker
        self.sizeScale = sizeScale
        self.overlappingDomains = overlapDoms

    def transformCoords(self, transf):
        if np.shape(transf) == (2, 2):
            self.rotBeams = np.dot(self.beams, transf)
        else:
            raise




"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes.woods ***

Defines the Woods class, used for conversion of a Woods' notation string into
a matrix and vice versa, as well as for formatting

Author: Michele Riva
Created: 2021-03-13
"""

import re
import itertools

import numpy as np


# TODO: fix method names
# TODO: allow also "rect" special Woods notation for hexagonal lattices (ONLY??)

degrees = '\u00b0'

class Woods:
    common = {'p(1\u00d71)', 'p(1\u00d72)', 'p(2\u00d71)', 'p(2\u00d72)',
              'p(3\u00d71)', 'p(3\u00d72)', 'p(3\u00d73)', 'c(2\u00d72)',
              'c(4\u00d74)', 'c(6\u00d72)', 'c(8\u00d72)'}
    examples = {
        'Oblique': common,
        'Square': common | {'c(4\u00d72)',
                            'p(\u221a2\u00d7\u221a2)R45' + degrees,
                            'p(\u221a5\u00d7\u221a5)R26.6' + degrees,
                            'p(2\u221a2\u00d7\u221a2)R45' + degrees,
                            'c(3\u221a2\u00d7\u221a2)R45' + degrees,
                            'c(5\u221a2\u00d7\u221a2)R45' + degrees},
        'Rectangular': common,
        'Hexagonal': common | {'c(4\u00d72)',
                               'p(\u221a3\u00d7\u221a3)R30' + degrees,
                               'p(\u221a7\u00d7\u221a7)R19.1' + degrees,
                               'p(2\u221a3\u00d72\u221a3)R30' + degrees},
        'Rhombic': common,
        }
    
    def woodsToMatrix(self, woods, bulk):
        parsed = self.parseWoods(woods)
        if parsed is None:
            return None
        g1 = parsed[1]
        g2 = parsed[2]
        alpha = np.radians(parsed[3])
        R1 = np.linalg.norm(bulk[0])
        R2 = np.linalg.norm(bulk[1])
        q = R2/R1
        omega = np.arccos(np.dot(bulk[0], bulk[1])/(R1*R2))
        
        m = np.array([[g1*np.sin(omega-alpha), g1*np.sin(alpha)/q],
                      [-g2*q*np.sin(alpha), g2*np.sin(omega+alpha)]])
        m = m/np.sin(omega)
        
        if parsed[0] == 'c':
            m = np.dot([[1, 1], [-1, 1]], m)/2
        
        if self.isCommensurate(m):
            m = np.array([int(np.round(mij)) 
                          for mij in m.ravel()]).reshape(m.shape)
        else:
            m = None  # incommensurate
        return m
    
    def parseWoods(self, woods):
        if not isinstance(woods, str):
            return None
        
        reWoods = re.compile(
            r'''^(?P<prefix>[pc])                # * primitive or centered
            \(                                   # * open parenthesis
            (?P<g1>\d+)?                         # * direction1, integer part
            (\u221a(?P<g1rt>\d+))?               # * direction1, radical part
            \u00d7                               # * times
            (?P<g2>\d+)?                         # * direction2, integer part
            (\u221a(?P<g2rt>\d+))?               # * direction2, radical part
            \)                                   # * close parenthesis
            ((R(?P<alpha>\d+(\.\d+)?))\u00b0)?$  # * rotation angle
            ''', re.VERBOSE)
        # notice that the ^ and $ anchors make sure that the string is 
        # matched as a whole
        
        # check if it matches the full woods
        m = reWoods.match(woods)
        if m is not None:
            w = m.groupdict()
            g1 = 1
            g2 = 1
            if w['g1'] is not None:
                g1 *= float(w['g1'])
            if w['g1rt'] is not None:
                g1 *= np.sqrt(float(w['g1rt']))
            if w['g2'] is not None:
                g2 *= float(w['g2'])
            if w['g2rt'] is not None:
                g2 *= np.sqrt(float(w['g2rt']))
            if w['alpha'] is not None:
                alpha = float(w['alpha'])
            else:
                alpha = 0
            return [w['prefix'], g1, g2, alpha]
        return None
    
    def matrixToWoods(self, m, bulk):
        m = np.array(m)
        
        pc = self.primitiveOrCentered(m, bulk)
        if pc:  # matrix is woods representable as primitive or centered
            toFormat = []
            for gSq in [pc[1]**2, pc[2]**2]:
                gSq = np.round(gSq)
                (g2, grt) = self.squareToProdOfSquares(gSq)
                integer = str(int(round(np.sqrt(g2))))
                
                dirStr = '' #format the direction in here
                if grt > 1: # insert root part
                    dirStr = '\u221a' + str(int(grt))
                if dirStr == '': # if there is no root part, always insert 
                                 # the integer part
                    dirStr = integer
                else:
                    if integer != '1': # otherwise place it in only if it's 
                                       # not 1
                        dirStr = integer + dirStr
                
                toFormat.append(dirStr)
            woods = '{}({})'.format(pc[0], '\u00d7'.join(toFormat))
            cosAlpha = pc[3]
            if np.abs(cosAlpha) > 1e-3 and 1-np.abs(cosAlpha) > 1e-3:
                #angle not 0, 90 nor 180
                alpha = np.round(np.degrees(np.arccos(cosAlpha)),
                                 decimals = 1)
                woods += 'R' + str(alpha) + degrees
            return woods
        return None
    
    def isCommensurate(self, m, eps=1e-3):
        if m is None:
            return False
        
        mu = np.linalg.det(m)
        if np.round(mu) == 0.0:  # matrix is singular
            return False

        if np.abs(mu/np.round(mu) - 1) > eps:  # determinant is not integer
            return False
        
        # now check whether any element is non-integer
        for mij in m.ravel():
            if np.round(mij) == 0.0:
                if abs(mij)/np.sqrt(np.abs(mu)) > eps:
                    return False
            elif np.abs(mij/np.round(mij) - 1) > eps:
                return False
        return True
    
    def isRepresentable(self, m, basis):
        transform = np.dot(m, basis)
        nBasis = np.linalg.norm(basis, axis=1)
        nTransf = np.linalg.norm(transform, axis=1)
        g1 = nTransf[0]/nBasis[0]
        g2 = nTransf[1]/nBasis[1]
        mu = np.abs(np.linalg.det(m))
        
        return abs(mu/(g1*g2) - 1) < 1e-8
    
    def primitiveOrCentered(self, m, basis):
        invT = np.array([[1, -1], [1, 1]])
        
        primitive = self.isRepresentable(m, basis)
        ctrd = self.isRepresentable(np.dot(invT, m), basis)
        if primitive:
            prefix = 'p'
        elif ctrd:
            prefix = 'c'
            m = np.dot(invT, m)
        else:
            return False
        
        nBasis = np.linalg.norm(basis, axis=1)
        transform = np.dot(m, basis)
        nTransf = np.linalg.norm(transform, axis=1)
        g1 = nTransf[0]/nBasis[0]
        g2 = nTransf[1]/nBasis[1]
        cosAlpha = np.dot(transform[0], basis[0])/(nTransf[0]*nBasis[0])
        
        return (prefix, g1, g2, cosAlpha)
    
    def squareToProdOfSquares(self, number):
        # takes number, finds all prime factors, returns a tuple, first 
        # element is a product of all primes showing up an even number of 
        # times, the second one the rest. Useful to turn, e.g., sqrt(12) 
        # into 2sqrt(3)
        
        factors = list(self.primeFactors(number))
        if not factors:
            factors = [1]
        uniqueFact = sorted(list(set(factors)))
        countFact = [((factors.count(fac) // 2)*2, factors.count(fac) % 2)
                      for fac in uniqueFact]
        (pow2, restPow) = zip(*countFact)
        (squares, remainders) = zip(*[(fact**pow, fact**rem)
                                      for (fact, pow, rem)
                                      in zip(uniqueFact, pow2, restPow)
                                      ])
        
        return (np.prod(squares), np.prod(remainders))
    
    def primeFactors(self, n):  # TODO: get a reference for this code
        f = 2
        increments = itertools.chain(
                                  [1, 2, 2],
                                  itertools.cycle([4, 2, 4, 2, 4, 6, 2, 6])
                                  )
        for incr in increments:
            if f*f > n:
                break
            while n % f == 0:
                yield f
                n //= f
            f += incr
        if n > 1:
            yield n



import time
import timeit # dependency for testing! Don't use in Viperleed build

import numpy as np

from rfactor import r_factor_new as rf

"""
For compiling:
f2py -m rfactor rfactor.f90 -h rfactor.pyf --overwrite-signature
f2py -c rfactor.pyf rfactor.f90 
"""

# Prepare some example arrays with curve data

y1 = np.array([-1, -0.5 ,0, 0.5, 1, 1.5, 2.5, 1.9,1,0])
y2_match = np.array([-1, -0.5 ,0, 0.5, 1, 1.5, 2.5, 1.9,1,0])
y2_bad_match = np.array([-1, -1.5 ,-1, -0.5, 0.5, 1.5, 3.5, 1.9,-2,0])

E_start1 = 20.0
E_start2 = 20.0
E_step = 1.0

V0r_shift = 0

R_pendry = None
numerator = None
denominator = None
N_overlapping_points = None

# Test calls
print("Test exec:")
print(rf.test_exec(1.0))
print("\n")

R_pendry, numerator, denominator, N_overlapping_points = rf.r_factor_beam(y1, y2_match, E_start1, E_start2, E_step, V0r_shift)

print("Matching: R, num, dem, N_ovl")
print(R_pendry)
print(numerator)
print(denominator)
print(N_overlapping_points)
print("\n")

R_pendry, numerator, denominator, N_overlapping_points = rf.r_factor_beam(y1, y2_bad_match, E_start1, E_start2, E_step, V0r_shift)

print("Non-Matching: R, num, dem, N_ovl")
print(R_pendry)
print(numerator)
print(denominator)
print(N_overlapping_points)
print("\n")

########
print("\n")
print("Trapez Integration Test:")
dx = 1
array = np.arange(-67, 185, step=dx)
print("Expected result: " + str(np.trapz(array)))
result = rf.trapez_integration_const_dx(array, dx)
print("Result: " + str(result))

############################################
#  BEAMSET
############################################

y1 = np.array([y1,y1,y1])
y2 = np.array([y2_match,y2_bad_match, y2_match])
sizes_y1 = [10,10,10]
sizes_y2 = [10,10,10]
E_start1 = np.array([20,20,20])
E_start2 = np.array([20,18,20])
E_step = 1
V0r_shift = 0
beamtypes = np.array([0,0,0])

R_Pe_weighted, R_Pe_beams, N_overlapping_points = rf.rfactor_beamset(y1, sizes_y1, y2, sizes_y2, E_start1, E_start2, E_step, V0r_shift, beamtypes)

print("Beamset Test:")
print(R_Pe_weighted)
print(R_Pe_beams)
print(N_overlapping_points)


"""Handling of TensErLEED Fortran source code files.

Checks for the presence of the TensErLEED source code in a given path,
deals with versioning, and provides the path to the TensErLEED source
code.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-04-09'
__license__ = 'GPLv3+'

import logging
import os
from pathlib import Path
import zipfile

from viperleed import VIPERLEED_TENSORLEED_ENV
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)

TENSERLEED_FOLDER_NAME = 'TensErLEED'

# TensErLEED Versions
KNOWN_TL_VERSIONS = (
    '1.6.0',
    '1.6.1',
    '1.7.1',
    '1.7.2',
    '1.7.3',
    '1.7.4',
    '1.7.5',
    '1.7.6',
    '2.0.0',
    )

OLD_TL_VERSION_NAMES = {
    '1.6': '1.6.0',
    '1.61': '1.6.1',
    '1.71': '1.7.1',
    '1.72': '1.7.2',
    '1.73': '1.7.3',
    '1.74': '1.7.4',
    '1.75': '1.7.5',
    '1.76': '1.7.6',
    }

CURRENT_TL_VERSION = max([Version(v) for v in KNOWN_TL_VERSIONS])

VERSION_FILE_NAME = 'version'

_KNOWN_TENSORLEED_TOP_FOLDERS = {
    'viperleed-tensorleed',
    'tensorleed',  # For backwards compatibility
    }

def get_tensorleed_path(tensorleed_path=None):
    """Return the path to the TensErLEED source code.

    Parameters
    ----------
    tensorleed_path : pathlike, optional
        Path to the viperleed-tensorleed source code, by default None.
        If not given, the $VIPERLEED_TENSORLEED environment variable
        is used, if defined. As a last resort, the current directory
        is used. In all cases, the path given may be the top-level tree
        containing the tensor-LEED code, its parent, or a first-level
        subfolder.

    Returns
    -------
    Path
        Path to the TensErLEED source code, that is, the directory
        that contains, e.g., potentially multiple TensErLEED sources
        as well as the EEASiSSS source code and compiled binaries.

    Raises
    ------
    ValueError
        If neither the tensorleed_path argument nor the
        VIPERLEED_TENSORLEED environment variable are set.
    FileNotFoundError
        If neither tensorleed_path (or VIPERLEED_TENSORLEED), its
        parent, or any of its children, point to one of the known
        tensor-LEED directory structures.
    """
    if tensorleed_path is not None:
        path_ = Path(tensorleed_path)
        return _verify_tensorleed_path(path_)

    # Check environment variable $VIPERLEED_TENSORLEED
    try:
        path_ = Path(os.environ[VIPERLEED_TENSORLEED_ENV])
    except KeyError:  # Environment variable not set. Try CWD below.
        pass
    else:
        return _verify_tensorleed_path(path_)

    # Last resort: try with the current directory
    try:
        return _verify_tensorleed_path(Path.cwd())
    except ValueError:
        raise ValueError(
            'TensErLEED path not specified.\n'
            'Either pass a path to the TensErLEED source code with the '
            '--tensorleed command-line argument, or set the environment '
            f'variable {VIPERLEED_TENSORLEED_ENV}.'
            ) from None


def _verify_tensorleed_path(path_):
    """Return the path to the tensor-LEED folder, starting from `path_`.

    Parameters
    ----------
    path_ : Path
        Path where the tensor-LEED code is looked up.

    Returns
    -------
    verified_path : Path
        Path to the folder that contains tensor-LEED code. The
        following places are checked: `path_`, its containing
        folder, and all of its direct subfolders. `verified_path`
        is identified as a tensor-LEED containing folder if its
        name is one of the known tensor-LEED repository names
        or if it contains any subfolder or zip file whose name starts with
        'TensErLEED'.

    Raises
    ------
    FileNotFoundError
        If `path_` is not the path to an existing directory,
        or if finding a verified tensor-LEED path failed.
    ValueError
        If the version of the TensErLEED source code can not be determined.
    """
    source = path_.resolve()
    if not source.is_dir():
        raise FileNotFoundError(f'Tensor-LEED directory {path_} not found.')

    # Collect various potential source-code places: the given path_,
    # its containing folder, and all of the path_'s direct subfolders
    potential_sources = (
        source,
        source.parent,
        *(d for d in source.iterdir() if d.is_dir())
        )

    # First look for an exact name match with
    # the known top-level tensor-LEED folders
    try:
        source_by_name = next(d for d in potential_sources
                              if d.name in _KNOWN_TENSORLEED_TOP_FOLDERS)
    except StopIteration:  # Not a known folder name
        pass
    else:
        if source_by_name != source:
            logger.warning(
                f'{source_by_name.name!r} directory found in '
                f'{source_by_name.parent}. Using {source_by_name} '
                f'instead of {source}.'
                )
        return source_by_name

    # Not found by name. Try to find if one contains a
    # TensErLEED directory instead. In principle here we
    # could use Rparams.get_tenserleed_directory() but we
    # do not care about versions and globbing is enough.
    try:
        with_tenserleed = next(
            d for d in potential_sources
            if any(d.glob(f'{TENSERLEED_FOLDER_NAME}*'))
            )
    except StopIteration:
        raise FileNotFoundError('Could not find a known tensor-LEED '
                                f'source directory at {source}.') from None
    if with_tenserleed != source:
        logger.warning(
            f'TensErLEED code found in {with_tenserleed}. '
            f'Using the latter instead of {source}.'
            )
    return with_tenserleed


class TensErLEEDSource:

    def __init__(self, path):
        self.path = Path(path).resolve()
        # Check that the path is a directory or a zip file
        if self.path.is_dir():
            self.is_zipped = False
        elif self.path.is_file() and self.path.suffix == '.zip':
            self.is_zipped = True
        else:
            raise FileNotFoundError(
                f'TensErLEED path {self.path} is not a directory or a zip file')
        # make sure we can determine a version
        try:
            self.version = self._detect_version()
        except ValueError:
            msg = ("Unable to detect Version for TensErLEED source at "
                   f"{self.path}.")
            logger.error(msg)
            raise ValueError(msg)
        if self.is_zipped:
            # we don't support zipped source code yet (Issue #34).
            raise NotImplementedError(
                f"Please unzip {self.path} before using it.")

    def _detect_version(self):
        """Return the version of the TensErLEED source code.

        Returns
        -------
        Version
            Version of the TensErLEED source code.
        """
        # check if "-v" is in the name
        if '-v' in self.path.name:
            version_str = self.path.name.split('-v')[-1]
            version_str = version_str.replace('.zip','')
            version_str = OLD_TL_VERSION_NAMES.get(version_str, version_str)
            return Version(version_str)
        elif (self.path / VERSION_FILE_NAME).is_file():
            return version_from_version_file(self.path / VERSION_FILE_NAME)
        else:
            raise ValueError(f"TensErLEED source at {self.path} is missing"
                             "a version identifier.")


def version_from_version_file(path):
    """Return the version of the TensErLEED source code from a version file."""
    # if passed a dir, try to find the version file
    if path.is_dir():
        version_file = path / VERSION_FILE_NAME
        if not version_file.is_file():
            raise FileNotFoundError(f'No valid version file found in {path}.')
    else:
        version_file = path
    version_str = version_file.read_text().strip()
    return Version(version_str.replace('v', ''))


def get_tenserleed_sources(path):
    path = Path(path)
    sources = []
    for directory in path.glob(f'*{TENSERLEED_FOLDER_NAME}*'):
        sources.append(TensErLEEDSource(directory))
    if not sources:
        raise FileNotFoundError('No TensErLEED source directories found in '
                                f'{path}.')
    return sources



"""Package beams of viperleed.utilities.

Collects utilities for converting beams in different formats to the
EXPBEAMS.csv that is used in ViPErLEED.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2024-03-27'
__license__ = 'GPLv3+'

EXPBEAMS_DEFAULT = 'EXPBEAMS.csv'



"""Package slab of viperleed.calc.classes.

This package is a result of the refactor of what used to be the slab
module. It is now split into multiple submodule to ease maintenance.
The API of the package remains unchanged.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-02-21'
__license__ = 'GPLv3+'

from .bulk_slab import BulkSlab
from .errors import AlreadyMinimalError
from .errors import AtomsTooCloseError
from .errors import EmptySlabError
from .errors import InvalidUnitCellError
from .errors import MissingBulkSlabError
from .errors import MissingElementsError
from .errors import MissingLayersError
from .errors import MissingSublayersError
from .errors import NoBulkRepeatError
from .errors import NoVacuumError
from .errors import NotEnoughVacuumError
from .errors import SlabError
from .errors import TooFewLayersError
from .errors import VacuumError
from .errors import WrongVacuumPositionError
from .surface_slab import SurfaceSlab as Slab



"""Moule defaults of viperleed.calc.classes.rparams.

Defines the default values of 'simple' user PARAMETERS. Not-so-simple
parameters, which are defined as their own classes in package special,
also take care of their own default values. This module was originally
(2019-06-13) part of the rparams.py module, refactored by Michele Riva
in Oct 2023.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-23'
__license__ = 'GPLv3+'

import logging


# The name we expect for folders containing TensErLEED Fortran code
TENSERLEED_FOLDER_NAME = 'TensErLEED'

# Notice that we cannot use a module-level global object(), as this
# module may be imported a number of times when using multiprocessing
NO_VALUE = None


# Notice that the defaults in here that may be mutated during execution
# are saved as immutable types to prevent inadvertent modification of
# this global, and are rather converted to their mutable equivalent
# in the relevant places. The only difference is dictionaries. Copies
# are used for them.
DEFAULTS = {
    # BASIC PARAMETERS
    'FILAMENT_WF': {
        'lab6': 2.65,  # This is the default if nothing is given
        'w': 4.5,
        },
    # GAUSSIAN_WIDTH and GAUSSIAN_WIDTH_SCALING are set by parameter
    # SEARCH_CONVERGENCE
    'GAUSSIAN_WIDTH' : 0.1,
    'GAUSSIAN_WIDTH_SCALING' : 0.5,
    'LOG_LEVEL' : {
        NO_VALUE: logging.INFO,
        'debug': logging.DEBUG,
        'v' : 5, 'verbose' : 5,
        'vv' : 1, 'vverbose' : 1,
        },
    'OPTIMIZE': {  # settings for fd optimization
        'which': 'none',
        'step': 0.,
        'minpoints': 4,
        'maxpoints': 10,
        'convergence': 0.,
        'maxstep': 0.,
        },
    'PHASESHIFT_EPS': {
        'r': 0.1,
        'n': 0.05,
        'd': 0.02, # default value
        'f': 0.01,
        },
    'RUN': (0, 1, 2, 3),
    'SEARCH_EVAL_TIME': 60,  # time interval between reads of SD.TL,            # TODO: should be dynamic?
    'SEARCH_MAX_DGEN': {'all': 0, 'best': 0, 'dec': 100},
    'SYMMETRY_FIX': '',
    'THETA': 0,   # perpendicular incidence
    'TL_VERSION': None,
    'PHI': 0,     # not needed in case of perpendicular incidence
    'ZIP_COMPRESSION_LEVEL': 2,

    # SPECIAL PARAMETERS
    'IV_SHIFT_RANGE': (-3, 3, NO_VALUE),  # NO_VALUE step: from data
    'LAYER_CUTS': 'dz(1.2)',
    'LMAX': (NO_VALUE, NO_VALUE),
    'SEARCH_CULL': (0.1, 'genetic'),
    'SYMMETRY_EPS': 0.1,  # z always equal to in-plane
    'THEO_ENERGIES': (NO_VALUE, NO_VALUE, NO_VALUE),
    'THEO_ENERGIES - no experiments': (20, 800, 3),
    }



"""Module read of viperleed.calc.files.parameters.

Initial version by @fkraushofer in 2020, major rewrite by @amimre
and @michele-riva in June 2023. This module used to be part of
parameters.py. Refactored in October 2023.

Functions for reading from a PARAMETERS file and for updating
an Rparams object at runtime from a user-modified PARAMETERS file.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-18'
__license__ = 'GPLv3+'

import logging
from pathlib import Path

from viperleed.calc.classes import rparams
from viperleed.calc.lib.base import parent_name

from .errors import MissingEqualsError
from .interpret import ParameterInterpreter
from .reader import ParametersReader
from .write import comment_out


_LOGGER = logging.getLogger(parent_name(__name__))


def read(filename='PARAMETERS'):
    """Return an Rparams with the raw contents read from a PARAMETERS file.

    Parameters
    ----------
    filename : str or Path, optional
        The file to be read. The default is 'PARAMETERS'.

    Returns
    -------
    rpars : Rparams
        Object storing parameters for current run. Contains the
        raw parameters read in this function in its `.readParams`
        attribute. The parameters read are not interpreted. For
        that, call `parameters.interpret` passing the the same
        `rpars` object.

    Raises
    ------
    FileNotFoundError
        If filename does not exist.
    ParameterNotRecognizedError
        If one of the parameters read from filename is not
        a known one, or if a parameter read has no value.
    """
    filename = Path(filename).resolve()
    if not filename.is_file():
        _LOGGER.error('PARAMETERS file not found.')
        raise FileNotFoundError(filename)

    rpars = rparams.Rparams()
    comment_out_stop = False
    with ParametersReader(filename, noisy=True) as param_file:
        while True:
            try:
                param, assignment = next(param_file)
            except StopIteration:
                break
            except MissingEqualsError as exc:
                _LOGGER.warning(exc)
                rpars.setHaltingLevel(2)
                continue
            if param == 'STOP':
                comment_out_stop = True
            else:
                rpars.readParams[param].append(assignment)

    if comment_out_stop:
        _LOGGER.warning(
            'PARAMETERS file: STOP was set at start of '
            'program. Modifying PARAMETERS to disable STOP; '
            're-insert it if you actually want to stop.'
            )
        comment_out(rpars, 'STOP', comment='Disabled at program start',
                    path=filename.parent, suppress_ori=True)
    return rpars


def update(rpars, filename='PARAMETERS', update_from=''):
    """Update `rpars` from file.

    The following parameters are considered:
    SEARCH_CONVERGENCE, STOP (and its legacy names).

    Parameters
    ----------
    rpars : Rparams
        Parameters for current run. Its attributes are updated
        if the contents of `filename` has changed.
    filename : str or Path, optional
        The file to be read. The default is 'PARAMETERS'.
    update_from : str or Path, optional
        Path to the directory in which to look for `filename`.
        The default is an empty string, corresponding to the
        current directory.

    Raises
    ------
    FileNotFoundError
        If no `filename` is found in `update_from`.
    ParameterError
        If one of the parameters to be updated has an invalid value.
    """
    filename = Path(update_from, filename).resolve()
    if not filename.is_file():
        _LOGGER.error('parameters.update: PARAMETERS file not found.')
        raise FileNotFoundError(filename)

    # Note that no slab is given to the interpreter. It is not needed
    # for STOP (+legacy names) and SEARCH_CONVERGENCE. Also, note that
    # we don't complain (again) about faulty parameters while reading
    interpreter = ParameterInterpreter(rpars)
    with ParametersReader(filename, noisy=False) as param_file:
        for param, assignment in param_file:
            if param == 'STOP':
                # pylint: disable=no-member
                # Method is added dynamically
                interpreter.interpret_stop(assignment)
            if rpars.STOP:
                return  # No need to continue reading
            if param == 'SEARCH_CONVERGENCE':
                interpreter.interpret_search_convergence(assignment=assignment,
                                                         is_updating=True)



"""Module __main__ of ViPErLEED (viperleed) calc.

See viperleed.calc.__init__.py for more information.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

import sys

from viperleed.calc.cli import ViPErLEEDCalcCLI

ViPErLEEDCalcCLI.run_as_script()



"""
======================================
  ViPErLEED Graphical User Interface
======================================
       *** module guilib ***

Created: 2020-01-10
Author: Michele Riva
       
"""

import os

from viperleed import GLOBALS

USE_GUI = GLOBALS['USE_GUI']

from viperleed.guilib.decorators import profile_calls, exec_time, profile_lines
from viperleed.guilib.helpers import (conventional_angles,
                                      two_by_n_array_to_tuples,
                                      two_by_two_array_to_tuple,
                                      remove_duplicates)
from viperleed.guilib.base import (get_equivalent_beams,
                                   project_to_first_domain, check_type,
                                   check_leed_params, check_multi_leed_params,
                                   catch_gui_crash, check_py_version,
                                   string_matrix_to_numpy,
                                   format_floats,  # probably not needed globally
                                   integer_part_length,
                                   parallel, orientation, screen_radius,
                                   BeamIndex, PlaneGroup, Lattice)

if (os.name == 'posix' and 'DISPLAY' not in os.environ.keys()) or not USE_GUI:
    # The environment does not have graphics capabilities.
    BACKEND = None
    GLOBALS['USE_GUI'] = False
else:
    # Import GUI modules
    if check_py_version('3.8', 'earlier'):
        BACKEND = 'mplcairo'
    else:
        BACKEND = 'agg'
    from viperleed.guilib.decorators import ensure_decorates_class
    from viperleed.guilib.widgetdecorators import (receive_mouse_broadcast,
                                                   broadcast_mouse)
    from viperleed.guilib.basewidgets import (Figure, FigureCanvas,
                                              MPLFigureCanvas, PainterMatrix,
                                              TextBox, TextBoxWithButtons)
    from viperleed.guilib.widgetslib import (AllGUIFonts, drawText,
                                             editStyleSheet,
                                             get_all_children_widgets)
    from viperleed.guilib.leedsim import *


from viperleed.guilib.leedsim.exportcsv import export_pattern_csv
from viperleed.guilib.leedsim.leedparameters import LEEDParameters
from viperleed.guilib.leedsim.leedparameters import LEEDParametersList
from viperleed.guilib.leedsim.classes import (LEEDPattern, LEEDSymmetryDomains,
                                              LEEDEquivalentBeams,)



"""Module files of viperleed.calc.

Contains functionality for reading and writing files. This
includes both input/output files for the whole viperleed.calc
as well as input/output files for the TensErLEED FORTRAN code.

Modules
-------
beamgen
    Functions to generate the BEAMLIST file, a TensErLEED input.
beams
    Functions for reading and writing various beams files.
displacements
    Functions for reading and interpreting the DISPLACEMENTS file.
iodeltas
    Reading and writing files for the delta-amplitudes calculation.
ioerrorcalc
    Reading and writing files for the error calculation.
iofdopt
    Functions for writing output from full-dynamic optimization.
iorefcalc
    Reading and writing files for the reference calculation.
iorfactor
    Reading and writing files for the R-factor calculation.
iosearch
    Reading, processing and writing files for the structure search.
iosuperpos
    Writing files relevant to the superpos calculation.
ivplot
    Functions for producing plots of I(V) curves.
patterninfo
    Functions for producing input files usable in the pattern
    simulator plug-in of the ViPErLEED GUI.
poscar
    Functions for reading and writing POSCAR files.
searchpdf
    Functions for writing SearchProgress/Report.pdf files.
vibrocc
    Functions for reading and writing the VIBROCC file.

Packages
--------
parameters
    Reading, writing, editing, and interpreting a PARAMETERS file.
"""
# Currently undocumented:
# delta_intensities
#     Too early, too messy.
# new_search
#     Too early, too messy.

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'



"""Module utils of viperleed.calc.files.parameters.

Initial version by @fkraushofer on 2020-08-18, major rewrite by
@amimre and @michele-riva in June 2023. This module used to be
part of parameters.py. Refactored in October 2023.

Contains functions and classes used in multiple submodules of
the viperleed.calc.files.parameters package.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-16'
__license__ = 'GPLv3+'

from collections.abc import Sequence
from dataclasses import dataclass, field


# TODO: some of these classes are probably also useful for other
# files, possibly with little modification. If they are, they
# could go higher up the hierarchy into a file.utils.py or similar


@dataclass(frozen=True)
class NumericBounds:
    """A container for bounds of numeric parameters.

    Attributes
    ----------
    type_ : {int, float}, optional
        Which numeric type these bounds are for.
    range : tuple, optional
        Minimum and maximum value for numeric parameter.
        Either limit can be None, signifying no limitation.
        Default is (None, None).
    accept_limits : tuple
        Whether the min/max endpoints of the range are
        acceptable values. Default is (True, True).
    out_of_range_event : {'fail', 'coerce', 'modulo'}, optional
        How to behave in case a number goes out of bounds.
        Default is 'fail'.
    """
    type_: type = float
    range_: tuple = (None, None)
    accept_limits: tuple = (True, True)
    out_of_range_event: str = 'fail'
    _msg_: str = field(init=False)

    def __post_init__(self):
        """Process assigned attributes.

        Raises
        ------
        ValueError
            If out_of_range_event is not one of the acceptable values
        ValueError
            If out_of_range_event is 'modulo' and some of the bounds
            were not specified
        ValueError
            If type_ is not one of the acceptable values
        """
        if self.out_of_range_event not in ('fail', 'modulo', 'coerce'):
            raise ValueError(
                f'Invalid out_of_range_event={self.out_of_range_event}. '
                'Must be "fail", "coerce", or "modulo"'
                )
        if self.type_ not in (int, float):
            raise ValueError('type_ must be int or float')
        is_modulo = self.out_of_range_event == 'modulo'
        if is_modulo and any(limit is None for limit in self.range_):
            raise ValueError('out_of_range_event "modulo" needs both limits')
        if is_modulo and not all(self.accept_limits):
            raise ValueError('out_of_range_event "modulo" needs both limits '
                             'to be acceptable.')

        # Use object.__setattr__ for frozen data-class
        object.__setattr__(self, '_msg_', self._make_message())

    @property
    def coerce(self):
        """Return whether the out-of-bounds reaction is 'coerce'."""
        return self.out_of_range_event == 'coerce'

    @property
    def fail(self):
        """Return whether the out-of-bounds reaction is 'fail'."""
        return self.out_of_range_event == 'fail'

    @property
    def max(self):
        """Return the lower bound."""
        return self.range_[1]

    @property
    def min(self):
        """Return the lower bound."""
        return self.range_[0]

    @property
    def closed_range(self):
        """A range within which values should fall in, including extremes."""
        increment = 1 if self.type_ is int else 1e-4  # float
        _min, _max = self.range_
        if _min is not None and not self.accept_limits[0]:
            _min += increment
        if _max is not None and not self.accept_limits[1]:
            _max -= increment
        return _min, _max

    @property
    def unlimited(self):
        """Return whether there is any bound."""
        return all(limit is None for limit in self.range_)

    def format_out_of_range(self, value):
        """Return a string for an out-of-range value."""
        return f'Value {value} is {self._msg_}'

    def is_in_range(self, val):
        """Return whether a value is within the acceptable bounds."""
        _min, _max = self.closed_range
        return (_min is None or val >= _min,
                _max is None or val <= _max,)

    def make_in_range(self, value):
        """Return a version of value that is within range."""
        in_range = self.is_in_range(value)
        if self.unlimited or all(in_range):
            return value
        if self.fail and not all(in_range):
            raise RuntimeError

        if self.out_of_range_event == 'modulo':
            _min, _max = self.range_
            return (value - _min) % (_max - _min) + _min

        assert self.coerce
        _min, _max = self.closed_range
        if _min is not None:
            value = max(value, _min)
        if _max is not None:
            value = min(value, _max)
        return value

    def _make_message(self):
        """Return a message for value-out-of-bounds."""
        if self.unlimited:
            return ''

        accept_min, accept_max = self.accept_limits
        range_chars = ({True: '[', False: ']'}, {True: ']', False: '['})
        if all(limit is not None for limit in self.range_):
            return (f'not in range {range_chars[0][accept_min]}{self.min}, '
                    f'{self.max}{range_chars[1][accept_max]}')
        if self.min is not None and accept_min:
            return f'less than {self.min}'
        if self.min is not None:
            return f'less than or equal to {self.min}'

        assert self.max is not None
        if accept_max:
            return f'larger than {self.max}'
        return f'larger than or equal to {self.max}'


POSITIVE_INT = NumericBounds(type_=int, range_=(1, None))
POSITIVE_FLOAT = NumericBounds(type_=float, range_=(0, None))


@dataclass(frozen=True)
class Assignment:
    """Class to store the flags and values of a parameter.

    Attributes
    ----------
    values_str : str
        The right-hand side of the parameter assignment. Can also be
        passed as a tuple of strings upon instantiation. In this case
        elements will be joined.
    flags_str : str
        The left-hand side of the parameter assignment, excluding the
        parameter itself. Can also be passed as a tuple of strings upon
        instantiation. In this case elements will be joined.
    parameter : str
        The PARAMETER this assignment corresponds to.
    flags : tuple
        The flags of the parameter as a tuple of strings.
        This is automatically generated from the string
        arguments given at instantiation.
    values : tuple
        The values of the parameter as a tuple of strings.
        This is automatically generated from the string
        arguments given at instantiation.
    """
    values_str: str
    parameter: str
    flags_str: str = ''
    flags: tuple = field(init=False)
    values: tuple = field(init=False)

    def __post_init__(self):
        """Split out left- and right-hand sides into flags and values."""
        try:
            object.__setattr__(self, 'parameter', self.parameter.strip())
        except AttributeError as exc:  # Not a string
            raise TypeError('parameter must be a string') from exc
        if not self.parameter:
            raise ValueError('parameter must contain printable characters')

        flags = self._unpack_assignment_side(self.flags_str)
        values = self._unpack_assignment_side(self.values_str)

        object.__setattr__(self, 'flags', flags)
        object.__setattr__(self, 'values', values)

        # Make sure values_str and flags_str are actually strings:
        # we also accept Sequence of strings at __init__
        if not isinstance(self.values_str, str):
            object.__setattr__(self, 'values_str', ' '.join(self.values_str))
        if not isinstance(self.flags_str, str):
            object.__setattr__(self, 'flags_str', ' '.join(self.flags_str))

    @property
    def flag(self):
        """Return the leftmost flag as a string."""
        return self.flags[0] if self.flags else ''

    @property
    def value(self):
        """Return the leftmost value as a string."""
        return self.values[0] if self.values else ''

    @property
    def other_flags(self):
        """Return all the flags except for the first one."""
        return self.flags[1:]

    @property
    def other_values(self):
        """Return all the values except for the first one (as strings)."""
        return self.values[1:]

    @staticmethod
    def _unpack_assignment_side(side):
        """Split the side left or right of the equal into bits."""
        if isinstance(side, str):
            return tuple(side.split())
        if isinstance(side, Sequence):
            return tuple(side)
        raise TypeError



"""ViPErLEED utility: Modify vacuum gap.

This utility takes a slab in POSCAR format and modifies the vacuum gap.
"""

__authors__ = (
    'Alexander M. Imre (@amimre)',
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-08-03'
__license__ = 'GPLv3+'

from copy import deepcopy
import logging

import numpy as np

from viperleed.calc.classes.slab.errors import (NotEnoughVacuumError,
                                                WrongVacuumPositionError)
from viperleed.utilities.poscar.base import _PoscarStreamCLI


logger = logging.getLogger(__name__)


def modify_vacuum(slab, vacuum_gap_size, absolute=False):
    """Modify the vacuum gap size of a slab.

    Parameters
    ----------
    slab : Slab
        Slab to prepare for relaxation,
    vacuum_gap_size : float
        Size of the vacuum gap in Angstroms. If absolute is True, this
        is the absolute size of the vacuum gap, else it is the amount
        of vacuum to add or remove from the slab.
    absolute : bool, optional
        If True, vacuum_gap_size is the absolute size of the vacuum gap.

    Returns
    -------
    Slab
        The slab with the modified vacuum gap.

    Raises
    ------
    RuntimeError
        If the resulting vacuum gap size is negative.
    """
    processed_slab = deepcopy(slab)
    processed_slab.check_a_b_in_plane()
    processed_slab.project_c_to_z()
    processed_slab.update_cartesian_from_fractional()

    slab_thickness = processed_slab.thickness
    current_gap_size = processed_slab.vacuum_gap

    vacuum_gap_size += current_gap_size if not absolute else 0

    # check there is enough space on top of the slab to reduce the vacuum gap
    if vacuum_gap_size < current_gap_size:
        top_atom_positions = max(at.cartpos[2] for at in slab)
        gap_on_top = slab.c_vector[2] - top_atom_positions
        if gap_on_top < (current_gap_size - vacuum_gap_size):
            raise RuntimeError('The resulting vacuum gap size would be larger '
                               'than the distance between the topmost atom and '
                               'the top of the cell.')

    logger.debug(f'Current vacuum gap size:\t{current_gap_size:9.3f}')
    logger.debug(f'New vacuum gap size:\t\t{vacuum_gap_size:9.3f}')

    if vacuum_gap_size < 0:
        raise RuntimeError('The resulting vacuum gap size would be negative.')

    new_c_vector = (
        processed_slab.c_vector
        / np.linalg.norm(processed_slab.c_vector)
        * (vacuum_gap_size + slab_thickness)
    )
    print(f'New c vector: {new_c_vector}')
    print(f'Old c vector: {processed_slab.c_vector}')

    processed_slab.c_vector[:] = new_c_vector
    processed_slab.collapse_cartesian_coordinates()
    processed_slab.update_fractional_from_cartesian()
    try:
        processed_slab.check_vacuum_gap()
    except NotEnoughVacuumError:
        raise RuntimeError('The resulting vacuum gap would be too small.')
    except WrongVacuumPositionError:
        raise RuntimeError('Cannot modify the vaccum gap as requested. Check '
                           'that there already is a vacuum gap in the POSCAR.')
    return processed_slab


class ModifyVacuumCLI(_PoscarStreamCLI, cli_name='modify_vacuum'):
    """Change the vacuum gap of a POSCAR."""

    long_name = 'modify vacuum gap'

    def add_parser_arguments(self, parser):
        """Add mandatory vacuum and optional --absolute arguments."""
        super().add_parser_arguments(parser)
        parser.add_argument(
            'vacuum',
            help=('Add (or remove, if negative) vacuum on top of '
                  'the slab. Value in angstrom. Default is zero.'),
            type=float,
            default=0.0,
            )
        parser.add_argument(
            '-a', '--absolute',
            help=('If set, the value given to vacuum is the absolute size '
                  'of the vacuum gap, else it is the amount of vacuum to add '
                  'or remove from the slab.'),
            action='store_true'
            )

    def parse_cli_args(self, args):
        """Check consistency of vacuum and absolute."""
        parsed_args = super().parse_cli_args(args)
        if parsed_args.absolute and parsed_args.vacuum < 0:
            self.parser.error('Absolute vacuum gap must be non-negative, '
                              f'but was {parsed_args.vacuum}.')
        return parsed_args

    def process_slab(self, slab, args):
        """Return a new slab with modified vacuum."""
        if args.absolute:
            logger.debug('Using absolute vacuum gap size.')
        try:
            return modify_vacuum(slab, args.vacuum, absolute=args.absolute)
        except RuntimeError as exc:
            self.parser.error(str(exc))
        return slab  # This is unreachable as parser.error sys-exits


if __name__ == '__main__':
    ModifyVacuumCLI.run_as_script()



"""Module atom_collections of viperleed.tleemlib.classes.

Contains definitions of collections (i.e., containers) of Atoms.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-06'
__license__ = 'GPLv3+'

from abc import abstractmethod
from collections.abc import Container, Iterable, MutableSequence, Sequence

from viperleed.calc.classes import atom


class AtomContainerError(Exception):
    """Base exception for containers of Atom objects."""


class AtomListOutOfDateError(AtomContainerError):
    """AtomList needs to update_atoms_map."""

    def __init__(self, *args):
        """Initialize instance."""
        if not args:
            args = ('Call .update_atoms_map',)
        super().__init__(*args)


class DuplicateAtomsError(AtomContainerError):
    """There are multiple atoms with the same num."""

    def __init__(self, *args):
        """Initialize instance."""
        if not args:
            args = ('Multiple atoms with the same .num',)
        super().__init__(*args)


# pylint: disable=too-few-public-methods
# Disabled because AtomCollection is essentially only meant to
# be the base class for collections of Atom objects. For now,
# the only reasonable methods are n_atoms and __contains__.
class AtomContainer(Container):
    """Base class for all other collections of Atom objects."""

    @property
    @abstractmethod
    def n_atoms(self):
        """Return the number of atoms in this AtomList."""
# pylint: enable=too-few-public-methods


class AtomList(AtomContainer, MutableSequence):
    """A sequence of Atom objects.

    Attributes
    ----------
    strict : bool
        Whether this AtomList only allows unique .num values. This
        attribute can temporarily be turned off if atom numbers are
        made unique later. After .num(s) are unique, it is advisable
        to restore strict to True and to call update_atoms_map.
        A strict AtomList has some of its operations optimized
        for O(1) access.
    """

    def __init__(self, *atoms, strict=True):
        """Initialize instance."""
        if atoms and len(atoms) == 1 and isinstance(atoms[0], Iterable):
            atoms = atoms[0]
        self._atoms = list(atoms)
        if not all(isinstance(at, atom.Atom) for at in self):
            raise TypeError(
                f'{type(self).__name__}: only Atom objects allowed'
                )
        self._map = {at.num: at for at in self._atoms}
        self.strict = strict  # True == allow only unique num
        self._sort_map = {}   # Atom: index, for restoring sorting

        if self.strict and len(self._map) != self.n_atoms:
            raise DuplicateAtomsError

    def __contains__(self, value):
        """Return whether an atom is in this list."""
        if not isinstance(value, atom.Atom):
            return False
        try:
            atom_in_self = self.get(value.num)
        except KeyError:
            return False
        return value is atom_in_self

    def __delitem__(self, index):
        """Remove atom(s) at index (or slice)."""
        try:
            to_remove = self._atoms[index]
        except IndexError:
            raise IndexError('list assignment index out of range') from None
        del self._atoms[index]
        if not isinstance(index, slice):
            to_remove = [to_remove]
        for atom_ in to_remove:
            del self._map[atom_.num]

    def __eq__(self, other):
        """Return whether this list is equal to another one."""
        try:
            equal = self.n_atoms == len(other)
        except TypeError:
            equal = False
        if not equal:
            return NotImplemented
        equal = all(at_self is at_other
                    for at_self, at_other in zip(self, other))
        if not equal:
            return NotImplemented
        return True

    def __iter__(self):
        """Return an iterator of atoms in this AtomList."""
        # Reimplemented as the Sequence ABC uses a for and __getitem__
        return iter(self._atoms)

    def __getitem__(self, index):
        """Return the Atom(s) at index (or slice)."""
        return self._atoms[index]

    def __len__(self):
        """Return the number of atoms in this list."""
        return len(self._atoms)

    def __repr__(self):
        """Return a string representation of this atom list."""
        _repr_ = f'{type(self).__name__}('
        _repr_ += ', '.join(str(at) for at in self)
        return _repr_ + f', strict={self.strict})'

    def __reversed__(self):
        """Return a reverse iterator of atoms in this AtomList."""
        return reversed(self._atoms)

    def __setitem__(self, index, value):
        """Assign the atom(s) at index (or slice) to value."""
        # When a slice, we expect an iterable, otherwise a single
        # element. For the case of a slice, make sure we consume a
        # possible iterator right away, as we'll need it twice
        if isinstance(value, Sequence):
            new_atoms = value
        elif isinstance(value, Iterable):
            new_atoms = tuple(value)
        elif not isinstance(index, slice):
            new_atoms = (value,)

        if isinstance(index, slice):
            replaced = set(self._atoms[index])
        else:
            replaced = {self._atoms[index]}

        try:
            _invalid = {at for at in new_atoms
                        if at not in replaced and at.num in self._map}
        except AttributeError:
            raise TypeError(
                f'{type(self).__name__}: only Atom objects allowed'
                ) from None
        if _invalid and self.strict:
            raise DuplicateAtomsError(
                f'Atom(s) {_invalid} have the same .num as others '
                f'already present in this {type(self).__name__}'
                )
        self._map.update((at.num, at) for at in new_atoms)
        if not isinstance(index, slice):
            new_atoms = new_atoms[0]
        self._atoms[index] = new_atoms

    @property
    def n_atoms(self):
        """Return the number of atoms in this AtomList."""
        return len(self)

    def clear(self):
        """Remove all atoms."""
        self._atoms.clear()
        self._map = {}
        self._sort_map = {}

    def copy(self):
        """Return a copy of this AtomList."""
        return self.__class__(*self, strict=self.strict)

    def count(self, value):
        """Return the number of occurrences of value."""
        try:
            present = self._map[value.num] is value
        except (KeyError, AttributeError):
            present = False
        if self.strict:
            return 1 if present else 0
        if not present:
            return 0
        return self._atoms.count(value)

    _no_default = object()

    def get(self, num, default=_no_default):
        """Return the atom with a given number."""
        if not self.strict:
            raise NotImplementedError(
                f'Cannot use .get for non-strict {type(self).__name__}. You '
                f'can use, e.g., next(at for at in atoms if at.num=={num}) '
                '(for the first occurrence)'
                )
        try:
            return self._map[num]
        except KeyError:
            pass
        if default is not self._no_default:
            return default
        if self.n_atoms != len(self._map):
            raise AtomListOutOfDateError
        raise KeyError(num)

    def index(self, value, *args):
        """Return the index of an atom."""
        return self._atoms.index(value, *args)

    def insert(self, index, value):
        """Insert a single atom at a given index."""
        if not isinstance(value, atom.Atom):
            raise TypeError(
                f'{type(self).__name__}: only Atom objects allowed'
                )
        if value.num in self._map and self.strict:
            raise DuplicateAtomsError(
                f'{value} has the same .num as antoher one '
                f'already present in this {type(self).__name__}'
                )
        self._atoms.insert(index, value)
        self._map[value.num] = value

    def restore_sorting(self, sort_map=None):
        """Restore a sorting order (or the last one stored)."""
        if not sort_map:
            sort_map = self._sort_map
        if not sort_map:
            return
        try:
            self.sort(key=sort_map.get)
        except (TypeError, KeyError) as exc:
            raise RuntimeError(
                'Impossible to restore sorting as some elements '
                'have been added since save_sorting() was called'
                ) from exc

    def reverse(self):
        """Reverse in place."""
        self._atoms.reverse()

    def save_sorting(self):
        """Store the current sorting so it can be later restored."""
        self._sort_map = dict(zip(self, range(self.n_atoms)))
        return self._sort_map

    def sort(self, *, key=None, reverse=False):
        """Sort this list of atoms."""
        self._atoms.sort(key=key, reverse=reverse)

    def update_atoms_map(self):
        """Refresh the {num: atom} map.

        Typically called after atom numbers have been modified.

        Raises
        ------
        DuplicateAtomsError
            If multiple atoms in self have the same .num.
        """
        self._map = {at.num: at for at in self._atoms}
        if len(self) != len(self._map):
            raise DuplicateAtomsError



"""Module sitetype of viperleed.calc.classes."""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2019-06-13'
__license__ = 'GPLv3+'

import logging
import re

import numpy as np

from viperleed.calc.lib.periodic_table import ATOMIC_MASS
from viperleed.calc.lib.periodic_table import COVALENT_RADIUS
from viperleed.calc.lib.periodic_table import PERIODIC_TABLE

logger = logging.getLogger(__name__)


class Sitetype:
    """Class storing properties of a site.

    Site types are identified by (main) element and name, and store
    vibrational amplitude and occupation"""

    def __init__(self, el, name):
        self.el = el.capitalize() # Element
        self.name = name
        self.label = self.el + '_' + self.name
        self.vibamp = {}    # vibrational amplitude per element
        self.occ = {}       # occupation per element

        self.oriState = None    # deep copy of self before a search is applied
        self.mixedEls = []      # stores the relevant rparams.ELEMENT_MIX

    def __str__(self):
        return self.label

    def isEquivalent(self, site2):
        """Checks whether two sites are equivalent, i.e. have the same label
        and the same values for vibrational amplitudes and occupations."""
        if (self.label == site2.label
                and self.vibamp == site2.vibamp
                and self.occ == site2.occ):
            return True
        return False

    def getVibAmp(self, rp, chemel):
        """Calculates a default vibrational amplitude for this site for the
        given chemical element from the element atomic mass, the experimental
        temperature and the material's Debye temperature."""
        if rp.T_DEBYE is None or rp.T_EXPERIMENT is None:
            logger.error("Cannot generate default vibrational amplitudes: "
                         "Temperature or Debye temperature undefined.")
            raise ValueError("Temperature and Debye temperature must be float")
        if chemel in rp.ELEMENT_RENAME:
            el = rp.ELEMENT_RENAME[chemel].capitalize()
        else:
            el = chemel.capitalize()
        if el not in PERIODIC_TABLE:
            logger.error(
                "Cannot generate default vibrational amplitude for site "
                + self.label + ": Element " + el + " not recognized.")
            raise ValueError("Element " + el + " not recognized.")
        if el not in ATOMIC_MASS:
            logger.error(
                "Cannot generate default vibrational amplitude for site "
                + self.label + ": Element" + el + " atomic mass unknown.")
            raise NotImplementedError("Element" + el + " atomic mass unknown.")
        scaling = 1.0
        for s in rp.VIBR_AMP_SCALE:
            try:
                label = s.split()[0]
                value = float(s.split()[1])
            except (ValueError, IndexError):
                logger.error("Failed to interpret VIBR_AMP_SCALE parameter "
                             "part, ignoring input: " + s)
                continue
            if value <= 0:
                logger.warning(
                    "VIBR_AMP_SCALE parameter: scaling values have "
                    "to be positive floats, ignoring input: " + s)
                continue
            mstr = re.escape(label)
            mstr = mstr.replace('\\*', '.*')
            m = re.match(mstr, self.label)
            if m:
                if m.end(0) == len(self.label):  # real match
                    scaling = value
        self.vibamp[chemel] = round(
            scaling * (np.sqrt(np.sqrt(1+16*((rp.T_EXPERIMENT
                                              / rp.T_DEBYE)**2))
                               * 109.15 / (ATOMIC_MASS[el]
                                           * rp.T_DEBYE))), 3)
        return


class Atom_type(Sitetype):
    """
    Collection of atoms with the same element and same site; inherited fromm Sitetype. Used for EEASiSSS input.
    """
    def __init__(self, el, name, new_bulk, layer = None):
        super().__init__(el, name)
        self.label = self.el + '_in_' + self.name
        self.covalent_radius = COVALENT_RADIUS[el.capitalize()]
        self.S = 0.32
        self.fxc = 1.0
        self.smallest_NN_dist = 1e10 # smallest NN distance of any atom of this type
        self.atoms = []
        self.bulk_layer = layer

        try:
            self.atomic_number = PERIODIC_TABLE.index(el.capitalize())+1
        except ValueError:
            logger.error('Invalid element symbol during Atom type assignment')
            raise ValueError('Invalid element symbol: ' + el.capitalize())

        self.new_bulk = new_bulk # attribute used in psgen; distinguishes if this atom is part of the auto-added bulk

    @property
    def getVibAmp(self, rp, chemel):
        raise AttributeError("Atom_type class objects are not intended to give a vibration amplitude. Use Site instead.")

    @property
    def isEquivalent(self, site2): #TODO: remove, not needed (I think)
        """Checks whether two sites are equivalent, i.e. have the same label
        and the same values for vibrational amplitudes and occupations."""
        if (self.label == site2.label
                and self.vibamp == site2.vibamp
                and self.occ == site2.occ
                and self.el == site2.el
                and self.new_bulk == site2.new_bulk):
            return True
        return False


    def set_MT_params(self, rmtmin, rmtmax, S):
        self.rmtmin = rmtmin
        self.rmtmax = rmtmax
        self.S = S

    def add_atom(self, atom, NN_dist):
        self.atoms.append(atom)
        if NN_dist < self.smallest_NN_dist:
            self.smallest_NN_dist = NN_dist


    def get_atomic_number(self):
        return self.atomic_number

    def get_type_NN_dist(self):
        return self.smallest_NN_dist

    def get_layer(self):
        return self.bulk_layer



"""Module energy_range of viperleed.calc.classes.rparams.special.

Classes
-------
EnergyRange
    Base class for parameters with 'start stop step' user input .
TheoEnergies(EnergyRange)
    Used as the Rparams attribute THEO_ENERGIES. The start -- stop
    intervalB is an integer multiple of step. All attributes are
    strictly positive.
IVShiftRange
    Used as the Rparams attribute IV_SHIFT_RANGE. Bounds are integer
    multiples of step. The step attribute is strictly positive.
"""

__authors__ = (
    'Michele Riva (@michele-riva)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2023-10-23'
__license__ = 'GPLv3+'

import ast
from collections.abc import Sequence
from dataclasses import dataclass
from decimal import Decimal
from math import ceil, floor, isfinite, remainder
from numbers import Real

from ..defaults import NO_VALUE
from .base import SpecialParameter


EPS = 1e-8  # Tolerance for comparisons of floats

# Notice that we purposely do not pass a param keyword here, as
# this is exclusively a base class, and it is not associated with
# any parameter. Hence, it is not registered by SpecialParameter.

@dataclass
class EnergyRange(SpecialParameter):
    """A container for energies."""

    start: float = NO_VALUE
    stop: float = NO_VALUE
    step: float = NO_VALUE

    def __eq__(self, other):
        """Return whether this EnergyRange is identical to another."""
        if not isinstance(other, (EnergyRange, Sequence)):
            return NotImplemented
        tuple_other = tuple(other)
        if not self.has_step and len(tuple_other) == 2:
            tuple_other = *tuple_other, NO_VALUE
        if len(tuple(self)) != len(tuple_other):
            return NotImplemented
        scale = abs(self.step) if self.has_step else 1
        for v_self, v_other in zip(self, tuple_other):
            if (v_self, v_other) == (NO_VALUE, NO_VALUE):
                continue
            if v_self is NO_VALUE or v_other is NO_VALUE:
                return NotImplemented
            if abs(v_self - v_other) > EPS*scale:
                return NotImplemented
        return True

    def __iter__(self):
        """Yield items of self."""
        yield self.start
        yield self.stop
        yield self.step

    def __post_init__(self):
        """Check and process initialization values."""
        non_defaults = self._non_defaults
        if not all(isinstance(e, Real) for e in non_defaults):
            raise TypeError('Values must be real')
        if not all(isfinite(e) for e in non_defaults):
            raise ValueError('All values must be finite')
        self._check_consistency()

        # Mess with start/stop only if all the values are present
        if self.defined:
            self.adjust_to_fit_step()

    @property
    def min(self):
        """Return the lower limit of this EnergyRange."""
        return self.start

    @property
    def max(self):
        """Return the upper limit of this EnergyRange."""
        return self.stop

    @property
    def defined(self):
        """Return whether all of the attributes have a value set."""
        return not any(v is NO_VALUE for v in self)

    @property
    def has_bounds(self):
        """Return whether both start and stop are defined."""
        return not any(v is NO_VALUE for v in (self.start, self.stop))

    @property
    def has_step(self):
        """Return whether this EnergyRange has step defined."""
        return self.step is not NO_VALUE

    @property
    def is_adjusted(self):
        """Return True, or raise if this EnergyRange is undefined."""
        if not self.defined:
            raise RuntimeError(f'{self} has undefined items')
        return True

    @property
    def n_energies(self):
        """Return the number of energies in this TheoEnergies."""
        if not self.defined:
            raise RuntimeError(f'{self} has undefined items')
        # +1 because we include both start and stop
        return round((self.stop - self.start) / self.step) + 1

    @property
    def _non_defaults(self):
        """Return the non-default values in self."""
        return [e for e in self if e is not NO_VALUE]

    def adjust_to_fit_step(self):
        """Adjust bounds to fit constraints with steps.

        The base-class implementation complains if this EnergyRange
        is not defined. This method should be extended in subclasses
        to fix inconsistencies of the bounds with the step.

        Raises
        ------
        RuntimeError
            If this method is called before all the `start`, `stop`
            and `step` attributes have been assigned.
        """
        if not self.defined:
            raise RuntimeError(f'{self} has undefined items')

    def contains(self, other, ignore_step=False):
        """Return whether other is a subset of this EnergyRange."""
        if not isinstance(other, EnergyRange):
            raise TypeError
        if not self.has_bounds:
            raise RuntimeError('Cannot compare non-defined TheoEnergies')
        if not other.has_bounds:
            raise ValueError('Cannot compare non-defined TheoEnergies')
        if self.start > other.start or self.stop < other.stop:
            return False
        if ignore_step:
            return True
        if not self.defined:
            raise RuntimeError('Cannot compare non-defined TheoEnergies')
        if not other.defined:
            raise ValueError('Cannot compare non-defined TheoEnergies')
        if self.step != other.step:
            return False
        # Finally, make sure they're not shifted
        self_shift = remainder(self.start, self.step)
        other_shift = remainder(other.start, other.step)
        return abs(self_shift - other_shift) < self.step * EPS

    def copy(self):
        """Return a (deep)copy of this EnergyRange."""
        return self.__class__(*self)

    @classmethod
    def from_value(cls, value):
        """Return an EnergyRange from value."""
        return cls(*value)

    @classmethod
    def from_sorted_grid(cls, energy_grid):
        """Return an energy range from a sorted grid of energies."""
        n_energies = len(energy_grid)
        if n_energies < 2:
            raise ValueError('Not enough energy_grid values. Need '
                             f'at least 2, found {n_energies}')
        start, stop = energy_grid[0], energy_grid[-1]
        step = (stop - start) / (n_energies - 1)
        return cls(start, stop, step)

    def is_equivalent(self, other):
        """Return whether self is equal to other, including swapping."""
        if self == other:
            return True
        # Try converting to an EnergyRange. Notice that we do not
        # use self.__class__ but always EnergyRange, as the base
        # class only swaps the bound values but never adjusts them
        try:
            range_other = EnergyRange(*other)
        except (ValueError, TypeError):
            return False
        return self == range_other

    def intersected(self, other):
        """Return the subset of this range in common with another.

        Parameters
        ----------
        other : EnergyRange
            The other range to intersect this one with.

        Returns
        -------
        intersected : EnergyRange
            The range common to self and other. The step of intersected
            is the same as the one of self. Notice that intersected may
            be wider than the actual common range. This may be the case
            if type(self) needs its bounds to somehow fit its step.

        Raises
        ------
        TypeError
            If other is not an EnergyRange
        ValueError
            If self and other have no energies in common, or if other
            has no bounds defined.
        RuntimeError
            If self has no bounds defined.
        """
        if not isinstance(other, EnergyRange):
            raise TypeError
        if not self.has_bounds:
            raise RuntimeError(f'{self} has no bounds')
        if not other.has_bounds:
            raise ValueError(f'{other} has no bounds')
        new_start = max(self.min, other.min)
        new_stop = min(self.max, other.max)
        if new_stop < new_start:
            raise ValueError('No intersection')
        return self.__class__(new_start, new_stop, self.step)

    @staticmethod
    def parse_string_sequence(string_sequence):
        """Return floating-point or NO_VALUE from a string sequence."""
        # We will use ast to interpret the sequence as a tuple.
        # Notice the trailing comma, in case string_sequence is
        # only one-element-long
        string = ','.join(string_sequence) + ','

        # We have to replace '_' with something that AST can
        # handle. 'None' seems easy enough. We replace it
        # again further down when converting the rest to float
        string = string.replace('_', 'None')
        try:
            return [float(v) if v is not None else NO_VALUE
                    for v in ast.literal_eval(string)]
        except (SyntaxError, ValueError, TypeError,
                MemoryError, RecursionError) as exc:
            new_exc = TypeError if 'float' in exc.args[0] else ValueError
            raise new_exc(' '.join(string_sequence)) from exc

    def set_undefined_values(self, *new_values):
        """Assign undefined values from new_values, then adjust start."""
        if len(new_values) == 1:
            new_values = new_values[0]
        for attr, value in zip(('start', 'stop', 'step'), new_values):
            if getattr(self, attr) is NO_VALUE:
                setattr(self, attr, value)
        self.__post_init__()  # Check and process new values

    def _check_consistency(self):
        """Change inconsistent values or complain."""
        try:
            1 / self.step
        except ZeroDivisionError:
            raise ValueError('Step cannot be zero') from None
        except TypeError:  # NO_VALUE
            pass

        start, stop, step = self
        if self.defined and (stop - start) * step < 0:
            raise ValueError('Inconsistent step. Cannot shift from '
                             f'{start:.2f} to {stop:.2f} with step={step:.2f}')
        if self.has_bounds and self._swap and stop < start:
            self._swap()

    def _swap(self):
        """Swap start and stop, change step."""
        self.start, self.stop = self.stop, self.start
        if self.has_step:
            self.step *= -1


class TheoEnergies(EnergyRange, param='THEO_ENERGIES'):
    """Energy range used for calculating I(V) curves."""

    @property
    def is_adjusted(self):
        """Return whether (stop - start) is a multiple of step."""
        if not self.defined:
            raise RuntimeError(f'{self} has undefined items')
        start, stop, step = self
        return abs(remainder(stop - start,  step)) < EPS

    def adjust_to_fit_step(self):
        """Modify start so that (stop - start) is a multiple of step."""
        # The next one raises RuntimeError if we're not fully defined:
        # We should not mess with the start/stop till we know all.
        if self.is_adjusted:
            return

        start, stop, step = self
        # The next line could in principle also be done with
        # remainder, but there are some corner cases in which
        # it is complicated to get the same results as now.
        start -= step - (stop - start) % step
        # if start < -EPS:    # Testing reveals that this is never hit
            # start = start % step
        if abs(start) < EPS:
            start = step
        self.start = start

    def as_floats(self):
        """Return a list of float values, replacing NO_VALUE with -1."""
        return [-1 if e is NO_VALUE else e for e in self]

    def expanded_by(self, n_steps):
        """Return a copy of this range wider by n_steps left and right."""
        if not self.defined:
            raise RuntimeError(f'{self} has undefined items')
        delta = n_steps * self.step
        start = max(self.step, self.min - delta)
        stop = self.max + delta
        if stop < start:
            raise ValueError(f'Cannot expand by n_steps={n_steps}. '
                             f'{self} would become empty')
        return self.__class__(start, stop, self.step)

    def _check_consistency(self):
        """Change inconsistent values or complain."""
        super()._check_consistency()
        if not all(e > 0 for e in self._non_defaults):
            raise ValueError('Values must be positive')


class IVShiftRange(EnergyRange, param='IV_SHIFT_RANGE'):
    """EnergyRange for Rparams attribute IV_SHIFT_RANGE."""

    @property
    def is_adjusted(self):
        """Return whether bounds are integer multiples of the step."""
        if not self.defined:
            raise RuntimeError(f'{self} has undefined items')
        *bounds, step = self
        return all(abs(remainder(b, step)) < EPS for b in bounds)

    @property
    def is_fixed(self):
        """Return whether this IVShiftRange has equal bounds."""
        if not self.has_bounds:
            raise RuntimeError(f'{self} has no bounds')
        scale = self.step if self.has_step else 1
        return abs(self.stop - self.start) < EPS * scale

    def adjust_to_fit_step(self):
        """Expand bounds so that they are integer multiples of step."""
        # The next one raises RuntimeError if we're not fully defined:
        # We should not mess with the start/stop till we know all.
        if self.is_adjusted:
            return

        # Make sure that adjusting does not un-fix a fixed range
        was_fixed = self.is_fixed

        # Exact floor/ceil comes from the 'exact' solution using
        # Decimal in https://stackoverflow.com/questions/28425705/
        start, stop, step = (Decimal(str(v)) for v in self)
        self.start = float(floor(start / step) * step)
        self.stop = float(ceil(stop / step) * step)

        if was_fixed and not self.is_fixed:
            raise RuntimeError(
                f'step={step!s} cannot be used to fix IV_SHIFT_RANGE to '
                f'{start}: {start} is not an integer multiple of {step}'
                )

    @classmethod
    def fixed(cls, fixed_value):
        """Return an IVShiftRange with both bounds at the same value."""
        return cls(fixed_value, fixed_value, NO_VALUE)

    def set_undefined_step(self, new_step):
        """Assign a new_step if this range has it undefined."""
        super().set_undefined_values(NO_VALUE, NO_VALUE, new_step)



"""Module iorefcalc of viperleed.calc.files.

Defines functions for reading and writing files relevant to the
reference calculation.
"""

__authors__ = (
    'Florian Kraushofer (@fkraushofer)',
    )
__copyright__ = 'Copyright (c) 2019-2024 ViPErLEED developers'
__created__ = '2020-08-19'
__license__ = 'GPLv3+'

import copy
import logging
import os

import fortranformat as ff
import numpy as np

from viperleed.calc.classes.beam import Beam
from viperleed.calc.files.beams import writeAUXBEAMS
from viperleed.calc.lib import leedbase
from viperleed.calc.lib.base import fortranContLine, splitMaxRight
from viperleed.calc.lib.version import Version

logger = logging.getLogger(__name__)


def combine_tensors(oripath=".", targetpath=".", buffer=0):
    """Combines Tensor files in 'oripath' into common files in 'targetpath'.
    The 'buffer' argument specified how much of a Tensor file can be stored in
    memory at a given time (in bytes). Set 0 to read the entire file at once
    (faster, but may be a problem if memory is limited)."""
    tensorfiles = [f for f in os.listdir(oripath) if f.startswith("T_")
                   and os.path.isfile(os.path.join(oripath, f))
                   and len(f.split("_")) == 3]
    tnum = {}
    for f in tensorfiles:
        try:
            num = int(f.split("_")[1])
            float(f.split("_")[2][:-2])
        except ValueError:
            continue   # not a real Tensor file
        if num not in tnum:
            tnum[num] = [f]
        else:
            tnum[num].append(f)
    if len(tnum) == 0:
        logger.error("Found no Tensor files to combine.")
        raise RuntimeError("No Tensor files found")
    for num in tnum:
        tlist = sorted(tnum[num], key=lambda x: float(x.split("_")[2][:-2]))
        name = os.path.join(targetpath, splitMaxRight(tlist[0], "_")[0])
        with open(name, "w") as wf:
            # having the 'if' this far outside duplicates code, but speeds
            #  up the loop
            if buffer == 0:
                for tf in tlist:
                    with open(os.path.join(oripath, tf), "r") as rf:
                        wf.write(rf.read())
                    try:
                        os.remove(os.path.join(oripath, tf))
                    except Exception:
                        logger.warning("Failed to delete file " + tf)
            else:
                for tf in tlist:
                    with open(os.path.join(oripath, tf), "r") as rf:
                        while True:
                            data = rf.read(buffer)
                            if data:
                                wf.write(data)
                            else:
                                break
                    try:
                        os.remove(os.path.join(oripath, tf))
                    except Exception:
                        logger.warning("Failed to delete file " + tf)
    return


def combine_fdout(oripath=".", targetpath="."):
    """Combines fd.out and amp.out files in oripath into a common file at
    targetpath."""
    for filetype in ("fd", "amp"):
        outlines = []
        startstr = filetype + "_"
        filelist = [f for f in os.listdir(oripath)
                    if os.path.isfile(os.path.join(oripath, f))
                    and f.startswith(startstr) and f.endswith("eV.out")]
        i = 0
        while i < len(filelist):
            try:
                float(filelist[i].split(startstr)[1].split("eV.out")[0])
                i += 1
            except ValueError:
                filelist.pop(i)
        if len(filelist) == 0:
            if filetype == "amp":
                return   # return without error, we don't need amp files
            logger.error("Found no fd.out files to combine.")
            raise RuntimeError("No fd.out files found")
        filelist.sort(key=lambda x: float(x.split(startstr)[1]
                                          .split("eV.out")[0]))
        nbeams = 0
        for f in filelist:
            with open(os.path.join(oripath, f), "r") as rf:
                lines = rf.readlines()
            lines = [s for s in lines if ".  CORRECT TERMINATION" not in s
                     and len(s.strip()) >= 1]
            if len(outlines) == 0:
                try:
                    nbeams = int(lines[1].strip().split()[0])
                except Exception:
                    logger.error("Failed to combine {}.out".format(filetype),
                                 exc_info=True)
                    raise
                outlines += lines
            else:
                outlines += lines[nbeams+2:]
            if (nbeams % 5) == 4:
                outlines += "\n"  # expects an empty line (fortran formatting)
            try:
                os.remove(os.path.join(oripath, f))
            except Exception:
                logger.warning("Failed to delete file " + f)
        try:
            with open(os.path.join(targetpath, (filetype + ".out")),
                      "w") as wf:
                wf.write("".join(outlines))
        except Exception:
            logger.error("Failed to write combined {}.out".format(filetype))
            raise
    return


def readFdOut(readfile="fd.out", for_error=False, ampfile="amp.out"):
    """Reads the fd.out file produced by the refcalc and returns a list of
    Beam objects. If 'ampfile' is set, will attempt to read complex amplitudes
    from the given file as well."""

    def parse_data(lines, which, filename, theobeams, nbeams):
        out_str = ""
        blocks = []
        for line in lines:
            if ".  CORRECT TERMINATION" in line:
                # sometimes superpos output is not in right line. sync issue?
                continue
            out_str += line + "\n"
            llist = line.split()
            if len(llist) == 0:
                continue  # skip empty lines
            try:
                float(llist[0])
            except ValueError:
                break  # end of data
            blocks.extend(llist)

        # Each block contains exactly nbeams+2 entries (double for amp)
        # reshape blocks so that each line corresponds to one block
        if which == "fd":
            blocks = np.reshape(blocks, (-1, nbeams+2))
        else:
            blocks = np.reshape(blocks, (-1, 2*nbeams+2))

        # and parse the data
        warned = False
        for block in blocks:
            if block[1] != "0.0001":
                if not for_error and not warned:
                    warned = True
                    logger.warning("File " + filename + "contains unexpected "
                                   "data for more than one structural "
                                   "variation. Only the first block was read.")
                continue
            en = float(block[0])
            values = [float(s) for s in block[2:]]
            for (j, beam) in enumerate(theobeams):
                if which == "fd":
                    beam.intens[en] = values[j]
                else:
                    beam.complex_amplitude[en] = complex(values[2*j],
                                                         values[2*j + 1])
        return out_str

    try:
        with open(readfile, 'r') as rf:
            filelines = [line[:-1] for line in rf.readlines()]
            rf.seek(0)
    except FileNotFoundError:
        logger.error("Error in readFdOut: file "+str(readfile)+" not found.")
        raise
    if filelines[0].startswith(" SOME ERROR OCCURED WHILE READING"):
        logger.error("File "+readfile+" reports error: "+filelines[0])
        raise Exception("File "+readfile+" reports error.")
    fdout = ""
    theobeams = []
    i = 1   # number lines as in text editor - be careful about indexing!
    nbeams = 1e10   # some large number, just to enter the while loop
    while i < nbeams+3:
        fdout += filelines[i-1] + "\n"
        llist = filelines[i-1].split()
        if i == 1:
            pass  # header - skip
        elif i == 2:
            nbeams = int(llist[0])
        else:
            theobeams.append(Beam((float(llist[1]), float(llist[2]))))
        i += 1

    # re-label the beams to get the correct number of characters and formatting
    mw = max([beam.getLabel()[1] for beam in theobeams])
    for beam in theobeams:
        beam.label = beam.getLabel(lwidth=mw)[0]

    fdout += parse_data(filelines[i-1:], "fd", readfile, theobeams ,nbeams)

    if ampfile and os.path.isfile(ampfile):
        with open(ampfile, 'r') as rf:
            amplines = [line[:-1] for line in rf.readlines()]
        # check header
        if any(amplines[i] != filelines[i] for i in range(1, nbeams + 2)):
            logger.warning("Failed to read " + ampfile + ": Header does not "
                           "match " + readfile)
            return theobeams, fdout
        # now read the rest
        parse_data(amplines[nbeams+2:], "amp", ampfile, theobeams, nbeams)
    if not theobeams:
        raise RuntimeError(f"No beams found in {readfile}")
    return theobeams, fdout


def writePARAM(sl, rp, lmax=-1):
    """Creats the contents of the PARAM file for the reference calculation.
    If no LMAX is passed, will use maximum LMAX from rp. Returns str."""
    if lmax == -1:
        lmax = rp.LMAX.max
    try:
        beamlist, beamblocks, beamN = writeAUXBEAMS(
            ivbeams=rp.ivbeams, beamlist=rp.beamlist, write=False)
    except Exception:
        logger.error("generatePARAM: Exception while getting data from "
                     "writeAUXBEAMS")
        raise

    # define Clebsh-Gordon coefficient tables:
    mnlmo = [1, 70, 264, 759, 1820, 3836, 7344, 13053, 21868, 34914, 53560,
             79443, 114492, 160952, 221408, 298809, 396492, 518206]
    mnlm = [1, 76, 284, 809, 1925, 4032, 7680, 13593, 22693, 36124, 55276,
            81809, 117677, 165152, 226848, 305745, 405213, 529036]

    # start generating output
    output = ('C  Dimension statements for Tensor LEED reference calculation, '
              '\nC  version v1.2\n\n')
    output += 'C  1. lattice symmetry\n\n'
    nl1, nl2 = leedbase.get_superlattice_repetitions(rp.SUPERLATTICE)
    ideg = 2  # any 2D point grid is at least 2fold symmetric
    # if sl.planegroup in ['p2','pmm','pmg','pgg','cmm','rcmm']:
    #     ideg = 2
    if sl.planegroup in ['p3', 'p3m1', 'p31m']:
        ideg = 3
    elif sl.planegroup in ['p4', 'p4m', 'p4g']:
        ideg = 4
    elif sl.planegroup in ['p6', 'p6m']:
        ideg = 3
        # should be 6, but according to TensErLEED fortran comments,
        #   3 works better
    output += ('      PARAMETER (MIDEG='+str(ideg)+',MNL1='+str(nl1)
               + ',MNL2='+str(nl2)+')\n')
    output += '      PARAMETER (MNL = MNL1*MNL2)\n'
    output += '\nC  2. General calculational quantities\n\n'
    output += '      PARAMETER (MKNBS = '+str(beamblocks)+')\n'
    output += '      PARAMETER (MKNT =  '+str(beamN)+')\n'
    output += ('      PARAMETER (MNPUN = '+str(len(beamlist))+', MNT0 = '
               + str(len(beamlist))+')\n')
    output += ('      PARAMETER (MNPSI = '+str(len(rp.phaseshifts))+', MNEL = '
               + str(len(rp.phaseshifts[0][1]))+')\n')
    output += '      PARAMETER (MLMAX = '+str(lmax)+')\n'
    output += ('      PARAMETER (MNLMO = '+str(mnlmo[lmax-1])+', MNLM = '
               + str(mnlm[lmax-1])+')\n')
    output += '\nC  3. Parameters for (3D) geometry within (2D) unit mesh\n\n'
    output += '      PARAMETER (MNSITE  = '+str(len(sl.sitelist))+')\n'
    output += '      PARAMETER (MNLTYPE = '+str(sl.n_layers)+')\n'
    mnbrav = 0
    mnsub = 0
    mnstack = 0
    if sl.bulkslab is None:
        sl.make_bulk_slab(rp)
    for layer in sl.non_bulk_layers:
        mnstack += 1
        if layer.n_atoms == 1:
            mnbrav += 1
        if layer.n_atoms > mnsub:
            mnsub = layer.n_atoms
    for i, layer in enumerate(sl.bulk_layers):
        if sl.bulkslab.layers[i].n_atoms == 1:
            mnbrav += 1
        if sl.bulkslab.layers[i].n_atoms > mnsub:
            mnsub = layer.n_atoms
    output += '      PARAMETER (MNBRAV  = '+str(mnbrav)+')\n'
    output += '      PARAMETER (MNSUB   = '+str(mnsub)+')\n'
    output += '      PARAMETER (MNSTACK = '+str(mnstack)+')\n'
    output += ('\nC  4. some derived quantities that must be treated '
               'explicitly (dummy dimensions for\n')
    output += 'C     special cases necessary\n\n'
    output += '      PARAMETER (MLMAX1=MLMAX+1)\n'
    output += '      PARAMETER (MLMMAX = MLMAX1*MLMAX1)\n\n'
    output += ('      PARAMETER (MNBRAV2 = '+('MNBRAV' if mnbrav > 0
                                              else '1')+')\n\n')
    output += ('      PARAMETER (MNCOMP= '+('MNLTYPE-MNBRAV' if mnsub > 1
                                            else '1 ')+')\n')
    output += ('      PARAMETER (MLMT  = '+('MNSUB*MLMMAX' if mnsub > 1
                                            else '1 ')+')\n')
    output += ('      PARAMETER (MNSUB2= '+('MNSUB * (MNSUB-1)/2' if mnsub > 1
                                            else '1 ')+')\n')
    output += ('      PARAMETER (MLMG  = '+('MNSUB2*MLMMAX*2' if mnsub > 1
                                            else '1 ')+')\n')
    output += ('      PARAMETER (MLMN  = '+('MNSUB * MLMMAX' if mnsub > 1
                                            else '1 ')+')\n')
    output += ('      PARAMETER (MLM2N = '+('2*MLMN' if mnsub > 1
                                            else '1 ')+')\n')
    output += ('      PARAMETER (MLMNI = '+('MNSUB*MLMMAX' if mnsub > 1
                                            else '1 ')+')\n')
    return output


def collectFIN(version=0.):
    """Combines AUXLATGEO, BEAMLIST, AUXNONSTRUCT, PHASESHIFTS, AUXBEAMS
    and AUXGEO into one string (input for refcalc), which it returns. Pass
    beamlist to avoid reading it again."""
    if version < Version('1.7.3'):
        filenames = ["AUXLATGEO", "BEAMLIST", "AUXNONSTRUCT", "PHASESHIFTS",
                     "AUXBEAMS", "AUXGEO"]
    else:
        filenames = ["AUXLATGEO", "BEAMLIST", "AUXNONSTRUCT", "PHASESHIFTS",
                     "AUXGEO"]
    fin = ""
    for fn in filenames:
        with open(fn, "r") as rf:
            fin += rf.read()
        if fin[-1] != "\n":
            fin += "\n"
    return fin


def writeAUXLATGEO(sl, rp):
    """Writes AUXLATGEO, which is part of the input FIN for the refcalc."""
    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'energies': ff.FortranRecordWriter('3F7.2'),
                     'uc': ff.FortranRecordWriter('2F7.4'),
                     'x_ase_wf': ff.FortranRecordWriter('2F7.4'),
                     }
        lj = 24  # ljust spacing
    else:
        formatter = {'energies': ff.FortranRecordWriter('3F9.2'),
                     'uc': ff.FortranRecordWriter('2F9.4'),
                     'x_ase_wf': ff.FortranRecordWriter('3F9.2'),
                     }
        lj = 30  # ljust spacing
    output = ''
    output += rp.systemName+' '+rp.timestamp+'\n'
    ens = [rp.THEO_ENERGIES.start,
           rp.THEO_ENERGIES.stop + 0.01,
           rp.THEO_ENERGIES.step]
    output += formatter['energies'].write(ens).ljust(lj) + 'EI,EF,DE\n'
    ucsurf = sl.ab_cell.T
    if sl.bulkslab is None:
        sl.make_bulk_slab(rp)
    ucbulk = sl.bulkslab.ab_cell.T
    output += formatter['uc'].write(ucbulk[0]).ljust(lj) + 'ARA1\n'
    output += formatter['uc'].write(ucbulk[1]).ljust(lj) + 'ARA2\n'
    if rp.TL_VERSION < Version('1.7.0'):
        output += ' 0.0    0.0             SS1\n'
        output += ' 0.0    0.0             SS2\n'
        output += ' 0.0    0.0             SS3\n'
        output += ' 0.0    0.0             SS4\n'
    output += formatter['uc'].write(ucsurf[0]).ljust(lj) + 'ARB1\n'
    output += formatter['uc'].write(ucsurf[1]).ljust(lj) + 'ARB2\n'
    if rp.TL_VERSION < Version('1.7.0'):
        output += ' 0.0    0.0             SO1\n'
        output += ' 0.0    0.0             SO2\n'
        output += ' 0.0    0.0             SO3\n'
        output += (formatter['x_ase_wf'].write([0.5, rp.V0_Z_ONSET]).ljust(lj)
                   + 'FR ASE\n')
    else:
        # Different parameters here in version 1.7! Previously in muftin
        output += (formatter['x_ase_wf'].write([rp.V0_IMAG, rp.V0_Z_ONSET,
                                                rp.FILAMENT_WF]).ljust(lj)
                   + 'V0i,ASE,WORKFN\n')
    try:
        with open('AUXLATGEO', 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write AUXLATGEO file")
        raise
    logger.debug("Wrote to AUXLATGEO successfully.")
    return


def writeAUXNONSTRUCT(sl, rp):
    """Writes AUXNONSTRUCT, which is part of the input FIN for the refcalc."""
    try:
        beamnums, _, _ = writeAUXBEAMS(ivbeams=rp.ivbeams,
                                       beamlist=rp.beamlist,
                                       write=False)
    except Exception:
        logger.error("generatePARAM: Exception while getting data from "
                     "writeAUXBEAMS")
        raise
    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'tst': ff.FortranRecordWriter('F7.4'),
                     'incidence': ff.FortranRecordWriter('(F7.2, F6.1)'),
                     'beamnums': ff.FortranRecordWriter('15I4'),
                     'eps': ff.FortranRecordWriter('F7.4'),
                     'ints': ff.FortranRecordWriter('I3'),
                     }
    else:
        formatter = {'tst': ff.FortranRecordWriter('F11.8'),
                     'incidence': ff.FortranRecordWriter('2F9.4'),
                     'beamnums': ff.FortranRecordWriter('15I5'),
                     'eps': ff.FortranRecordWriter('F7.4'),
                     'ints': ff.FortranRecordWriter('I3'),
                     }
    output = ''

    output += (formatter['tst'].write([rp.ATTENUATION_EPS]).ljust(18)
               + '>>>>> ! <<<<<              TST\n')
    output += formatter['beamnums'].write(beamnums)+'\n'
    output += (formatter['incidence'].write([rp.THETA, rp.PHI]).ljust(45)
               + 'THETA FI\n')
    output += formatter['eps'].write([rp.BULKDOUBLING_EPS]).ljust(45) + 'EPS\n'
    output += (formatter['ints'].write([rp.BULKDOUBLING_MAX]).ljust(45)
               + 'LITER\n')
    output += formatter['ints'].write([rp.LMAX.max]).ljust(45) + 'LMAX\n'
    if rp.TL_VERSION >= Version('1.7.0'):
        # TODO: if phaseshifts are calculated differently, change format here
        output += (formatter['ints'].write([1]).ljust(45)
                   + 'PSFORMAT  1: Rundgren_v1.6; 2: Rundgren_v1.7\n')
    if rp.TL_VERSION >= Version('1.7.3'):
        output += (formatter['ints'].write([1]).ljust(45)
                   + 'IFORM - formatted input and output\n')
    try:
        with open('AUXNONSTRUCT', 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write AUXNONSTRUCT file")
        raise
    logger.debug("Wrote to AUXNONSTRUCT successfully.")
    return


def writeAUXGEO(sl, rp):
    """Writes AUXGEO, which is part of the input FIN for the refcalc."""
    if rp.TL_VERSION < Version('1.7.0'):
        formatter = {'vibrocc': ff.FortranRecordWriter('2F7.4'),
                     'geo': ff.FortranRecordWriter('3F7.4'),
                     }
        lj = 26
    else:
        formatter = {'vibrocc': ff.FortranRecordWriter('2F9.4'),
                     'geo': ff.FortranRecordWriter('3F9.4'),
                     }
        lj = 32
    slab_c = sl.c_vector.copy()
    if rp.LAYER_STACK_VERTICAL:
        sl = copy.deepcopy(sl)
        sl.project_c_to_z()
        sl.update_layer_coordinates()
    output = ''
    output += ('---------------------------------------------------------'
               '----------\n')
    output += ('--- define chem. and vib. properties for different atomic '
               'sites ---\n')
    output += ('---------------------------------------------------------'
               '----------\n')
    i3 = ff.FortranRecordWriter('I3')
    ol = i3.write([len(sl.sitelist)]).ljust(lj)
    output += ol + 'NSITE: number of different site types\n'
    for i, site in enumerate(sl.sitelist):
        output += '-   site type '+str(i+1)+' ---\n'

        for el in sl.elements:
            # this reproduces the order of blocks contained in PHASESHIFTS:
            if el in rp.ELEMENT_MIX:
                chemelList = rp.ELEMENT_MIX[el]
            else:
                chemelList = [el]
            siteList = [s for s in sl.sitelist if s.el == el]
            for cel in chemelList:
                for s in siteList:
                    if s.isEquivalent(site):
                        occ, vib = site.occ[cel], site.vibamp[cel]
                        comment = ('Occ & VibAmp for '+cel+' in '+site.label
                                   + ' site')
                    else:
                        occ, vib = 0., 0.
                        comment = ''
                    try:
                        ol = formatter['vibrocc'].write([occ, vib]).ljust(lj)
                    except Exception:
                        logger.error(
                            "Exception while trying to write occupation / "
                            "vibrational amplitude for site " + site.label,
                            exc_info=True)
                    output += ol + comment + '\n'

    output += ('-----------------------------------------------------'
               '--------------\n')
    output += ('--- define different layer types                     '
               '           ---\n')
    output += ('-----------------------------------------------------'
               '--------------\n')
    ol = i3.write([sl.n_layers]).ljust(lj)
    output += ol + 'NLTYPE: number of different layer types\n'
    blayers = sl.bulk_layers
    layerOffsets = [np.zeros(3) for _ in range(sl.n_layers + 1)]
    if sl.bulkslab is None:
        sl.make_bulk_slab(rp)
    for i, layer in enumerate(sl.layers):
        output += '-   layer type '+str(i+1)+' ---\n'
        if layer.is_bulk:
            output += ('  2'.ljust(lj) + 'LAY = 2: layer type no. '
                       + str(i+1) + ' has bulk lateral periodicity\n')
        else:
            output += ('  1'.ljust(lj) + 'LAY = 1: layer type no. '
                       + str(i+1) + ' has overlayer lateral periodicity\n')
        if layer.is_bulk:
            bl = sl.bulkslab.layers[blayers.index(layer)]
            bulknums = {at.num for at in bl}
            bulkUnique = [at for at in layer if at.num in bulknums]
            natoms = len(bulkUnique)
            # sanity check: ratio of unit cell areas (given simply by
            #  SUPERLATTICE) should match ratio of written vs skipped atoms:
            arearatio = 1 / abs(np.linalg.det(rp.SUPERLATTICE))
            atomratio = len(bulkUnique) / layer.n_atoms
            if abs(arearatio - atomratio) > 1e-3:
                logger.warning(
                    'Ratio of bulk atoms inside/outside the bulk unit cell '
                    'differs from bulk/slab unit cell size ratio. This means '
                    'that the actual periodicity of the POSCAR does not match '
                    'the periodicity given in the SUPERLATTICE parameter. '
                    'Check SUPERLATTICE parameter and bulk symmetry!')
                rp.setHaltingLevel(2)
        else:
            natoms = layer.n_atoms
        ol = i3.write([natoms]).ljust(lj)
        output += ol+'number of Bravais sublayers in layer '+str(i+1)+'\n'
        if layer.is_bulk:
            writelist = bulkUnique
        else:
            writelist = layer.atlist
        writelist.sort(key=lambda atom: -atom.pos[2])
        for atom in writelist:
            writepos = atom.cartpos - atom.layer.cartori                        # TODO: .cartpos[2]. Issue #174
            ol = i3.write([sl.sitelist.index(atom.site)+1])
            if natoms != 1:
                ol += formatter['geo'].write([writepos[2],
                                              writepos[0], writepos[1]])
            else:
                # Bravais layers need to have coordinate (0., 0., 0.)
                #  -> store actual position for later, it will go into the
                #  interlayer vector
                ol += formatter['geo'].write([0., 0., 0.])
                layerOffsets[layer.num] += writepos
                layerOffsets[layer.num + 1] -= writepos
            ol = ol.ljust(lj)
            output += f'{ol}Atom N={atom.num} ({atom.el})\n'
    output += ('--------------------------------------------------------------'
               '-----\n')
    output += ('--- define bulk stacking sequence                             '
               '  ---\n')
    output += ('--------------------------------------------------------------'
               '-----\n')
    output += ('  0'.ljust(lj) + 'TSLAB = 0: compute bulk using layer '
               'doubling\n')

    # determine ASA
    if type(rp.BULK_REPEAT) == np.ndarray:
        bvectors_ASA = np.copy(rp.BULK_REPEAT) * np.array([1, 1, -1])
        if bvectors_ASA[2] < 0:
            bvectors_ASA = -bvectors_ASA
        bulkc = bvectors_ASA[2]
    else:
        bulkc = np.copy(rp.BULK_REPEAT)
        bvectors_ASA = -slab_c * bulkc/slab_c[2]
    # bulkc is now the repeat length along c. now correct for layer thickness:
    if rp.N_BULK_LAYERS == 2:
        bvectors_ASA[2] = bulkc - (blayers[1].cartbotz - blayers[0].cartori[2])  # TODO: .cartpos[2]. Issue #174
    else:
        bvectors_ASA[2] = bulkc - (blayers[0].cartbotz - blayers[0].cartori[2])  # TODO: .cartpos[2]. Issue #174

    # determine ASBULK - interlayer vector between bulk layers
    if rp.N_BULK_LAYERS == 2:
        # add layerOffsets for Bravais layers:
        bvectors_ASA += layerOffsets[blayers[0].num+1]
        # calculate ASBULK:
        bvectors_ASBULK = blayers[1].cartori - blayers[0].cartori               # TODO: .cartpos[2]. Issue #174
        bvectors_ASBULK[2] = blayers[1].cartori[2] - blayers[0].cartbotz
        bl2num = blayers[1].num
        # add layerOffsets for Bravais layers:
        bvectors_ASBULK -= layerOffsets[blayers[0].num+1]
        # correct ASA xy in case ASBULK already includes some:
        bvectors_ASA[:2] -= bvectors_ASBULK[:2]
    else:
        bl2num = blayers[0].num
        bvectors_ASBULK = bvectors_ASA

    ol = formatter['geo'].write([bvectors_ASA[2],
                                 bvectors_ASA[0], bvectors_ASA[1]]).ljust(lj)
    output += ol + 'ASA interlayer vector between different bulk units\n'
    ol = i3.write([blayers[0].num+1]).ljust(lj)
    output += (ol + 'top layer of bulk unit: layer type '+str(blayers[0].num+1)
               + '\n')
    ol = i3.write([bl2num+1]).ljust(lj)
    output += ol + 'bottom layer of bulk unit: layer type '+str(bl2num+1)+'\n'
    ol = formatter['geo'].write([bvectors_ASBULK[2], bvectors_ASBULK[0],
                                 bvectors_ASBULK[1]]).ljust(lj)
    output += (ol + 'ASBULK between the two bulk unit layers (may differ from '
               'ASA)\n')
    output += ('--------------------------------------------------------------'
               '-----\n')
    output += ('--- define layer stacking sequence and Tensor LEED output     '
               '  ---\n')
    output += ('--------------------------------------------------------------'
               '-----\n')
    nonbulk = sl.n_layers - rp.N_BULK_LAYERS
    if len(rp.TENSOR_OUTPUT) < nonbulk:    # check TENSOR_OUTPUT parameter
        if len(rp.TENSOR_OUTPUT) == 1:
            # interpret one value as concerning all
            rp.TENSOR_OUTPUT = rp.TENSOR_OUTPUT * nonbulk
        elif rp.TENSOR_OUTPUT:
            logger.warning(
                'Parameters TENSOR_OUTPUT is defined, but contains fewer '
                'values than there are non-bulk layers. Missing values '
                'will be set to 1.')
            rp.setHaltingLevel(1)
        for i in range(0, nonbulk-len(rp.TENSOR_OUTPUT)):
            rp.TENSOR_OUTPUT.append(1)
    if len(rp.TENSOR_OUTPUT) > nonbulk:
        logger.warning(
            'Parameters TENSOR_OUTPUT is defined, but contains more values '
            'than there are non-bulk layers. Excess values will be ignored.')
        rp.setHaltingLevel(1)
    ol = i3.write([sl.n_layers - rp.N_BULK_LAYERS]).ljust(lj)
    output += ol + 'NSTACK: number of layers stacked onto bulk\n'
    for layer in reversed(sl.non_bulk_layers):
        n = layer.num + 1
        v = sl.layers[n].cartori - layer.cartori
        v[2] = sl.layers[n].cartori[2] - layer.cartbotz                         # TODO: .cartpos[2]. Issue #174
        v = v + layerOffsets[n]   # add layerOffsets for Bravais layers
        ol = i3.write([n]) + formatter['geo'].write([v[2],
                                                     v[0], v[1]])
        output += (ol.ljust(lj) + 'layer '+str(n)+': layer type '+str(n)
                   + ', interlayer vector below\n')
        ol = i3.write([rp.TENSOR_OUTPUT[layer.num]]).ljust(lj)
        output += (ol + '0/1: Tensor output is required for this layer '
                   '(TENSOR_OUTPUT)\n')
        if rp.TENSOR_OUTPUT[layer.num] == 0:
            continue   # don't write the Tensor file names
        for i, atom in enumerate(layer):
            ol = f'T_{atom.num}'.ljust(lj)
            output += (ol + 'Tensor file name, current layer, sublayer '
                       + str(i+1) + '\n')
    output += ('--------------------------------------------------------------'
               '-----\n')
    output += ('--- end geometrical input                                     '
               '  ---\n')
    output += ('--------------------------------------------------------------'
               '-----\n')

    try:
        with open('AUXGEO', 'w') as wf:
            wf.write(output)
    except Exception:
        logger.error("Failed to write AUXGEO file")
        raise
    logger.debug("Wrote to AUXGEO successfully.")
    return


def writeMuftin(rp):
    """Writes a muftin.f file, which will be compiled for the refcalc."""
    output = """
C  Subroutine muftin contains explicit energy dependence of inner
C  potentials. The functional form should be left to the user entirely,
C  thus the subroutine is included in the input explicitly.

C  All quantities should be set in eV. Conversion to Hartree (atomic units)
C  is handled by ref-calc.f itself. Both the real and imaginary part should
C  be given as positive values (the program then handles the sign correctly).

      subroutine muftin(EEV,VO,VV,VPI,VPIS,VPIO)

C  global variable

C  EEV :  electron energy in the vacuum region
C  VO  :  difference between real part of inner potential in the bulk and in
C         topmost layer. Usually irrelevant -> set to zero.
C  VV  :  real part of the inner potential in the bulk.
C  VPI :  imaginary part of the inner potential.
C  VPIS:  in case different values for the imaginary part of the inner
C  VPIO:  potential for bulk and topmost layer were desired, VPIS would
C         correspond to the bulk value, VPIO would be the respective
C         value for the topmost layer. Usually, the effect is irrelevant,
C         i.e. both are set equal to VPI.

      real EEV,VO,VV,VPI,VPIS,VPIO

C  local variable

C  workfn: Work function of the LEED filament material. Theoretical predictions
C         for the inner potential usually have their energy zero fixed at the
C         Fermi level; the experimental energy scale (-> EEV) nominally
C         does the same, except for the fact that electrons need to overcome
C         the work function of the cathode first. Note that the this value is
C         thus formally determined by a LEED fit, yet don't trust its accuracy.

      real workfn

C  set work function of cathode
C  work function should be positive (added to exp. energy EEV)

      workfn = """
    output += str(round(rp.FILAMENT_WF, 4))+"\n"
    output += """
C  set real part of inner potential

"""
    if type(rp.V0_REAL) == list:
        oline = ("      VV = workfn-max({:.2f}, (({:.2f})+({:.2f})/sqrt("
                 "EEV+workfn+({:.2f}))))".format(*rp.V0_REAL))
    else:
        oline = "      VV = "+rp.V0_REAL
    output += fortranContLine(oline) + "\n"
    output += """
      write(6,*) workfn, EEV
      write(6,*) VV

c  set difference between bulk and overlayer inner potential

      VO = 0.

c  set imaginary part of inner potential - energy independent value used here

"""
    oline = "      VPI = "+str(rp.V0_IMAG)
    output += fortranContLine(oline) + "\n"
    output += """
C  set substrate / overlayer imaginary part of inner potential

"""
    oline = "      VPIS = "+str(rp.V0_IMAG)
    output += fortranContLine(oline) + "\n"
    oline = "      VPIO = "+str(rp.V0_IMAG)
    output += fortranContLine(oline) + "\n"
    output += """
      return
      end"""
    try:
        with open("muftin.f", "w") as wf:
            wf.write(output)
        logger.debug("Wrote to muftin.f successfully.")
    except Exception:
        logger.error("Exception while writing muftin.f file: ",
                     exc_info=True)
        raise
    return



"""
======================================
  ViPErLEED Graphical User Interface
======================================
 *** module guilib.leedsim.classes.leedpattern ***

Defines the LEEDPattern and LEEDsubpattern classes, used for displaying a
LEEDPattern. This is the NEW version of the classes, and replaces the old
implementations that is currently kept in oldleedpattern.py

Author: Michele Riva
Created: 2021-03-13
"""

import numpy as np

from viperleed import guilib as gl


class LEEDPattern:
    """
    LEEDPattern handles the LEED pattern of multiple structural domains, each of
    which may have multiple symmetry-induced domains. It also stores all the
    information needed to plot the LEED pattern.
    
    Parameters
    ----------
    It can be constructed from at least one instance of the following:
    - dict : assumed to be a LEED parameters dictionary
    - ConfigParser : assumed to be a LEED parameters dictionary
    - LEEDParameters
    - LEEDPattern
    - LEEDParametersList
    When passing multiple inputs, also an heterogeneous collection of the above
    is acceptable. However, the inputs will be checked for compatibility, i.e.,
    they should all have the same bulk lattice (both basis, group and 3D
    symmetry operations should match)
    
    duplicates : bool, optional (default=False)
        upon instantiation or when appending, determines whether or not to
        keep duplicates, i.e., input parameters that would generate identical
        LEED patterns
    """
    def __new__(cls, *args, **kwargs):
        """
        The only purpose of this is to handle the case in which a single
        LEEDPattern instance is passed. In that case the same instance is
        returned as is
        """
        if len(args) == 1 and isinstance(args[0], LEEDPattern):
            return args[0]
        return super().__new__(cls)
    
    # @gl.profile_calls(print_args=[30])
    def __init__(self, leed, *other_leeds, **kwargs):
        # Treat the special case of when a single LEEDPattern is passed:
        # __new__ already returns a reference to the LEEDPattern, so we
        # just skip everything
        if len(other_leeds) == 0 and isinstance(leed, LEEDPattern):
            return None

        # In all other cases, process the arguments one at a time,
        # collecting the necessary information. In fact, the only special
        # case to be treated is when one of the arguments is a LEEDPattern,
        # in which case we take its .parameters attribute (a LEEDParametersList)
        leeds = (leed, *other_leeds)
        for i, leed in enumerate(leeds):
            if isinstance(leed, LEEDPattern):
                leeds[i] = leeds[i].parameters

        # process acceptable keyword arguments
        duplicates = kwargs.get('duplicates', False)
        
        # and prepare the LEEDParametersList that defines the LEED pattern,
        # i.e., a list of STRUCTURAL domains. Consistency of bulk lattices
        # is delegated to LEEDParametersList
        self.__parameters = gl.LEEDParametersList(leeds, duplicates)

        # Construct the structural domains
        # TODO: self.domains = 
        

    @property
    def parameters(self):
        """
        Returns the LEEDParametersList that is used to build this LEED pattern
        """
        return self.__parameters

    @property
    def max_energy(self):
        """
        Maximum LEED energy
        """
        return self.__parameters['eMax']

    @property
    def n_domains(self):
        """
        Number of distinct domains produced by the symmetry operations of the
        bulk
        """
        return len(self.domains['operations'])

    @property
    def superlattices(self):                                                    # TODO: REMOVE?
        """
        Numpy ndarray of superlattice matrices generating the symmetry-related
        domains. The first element is the one generating the lattice whose basis
        is given in the constructor as a LEED parameter
        """
        superlattice = self.__parameters['SUPERLATTICE']
        operations = self.domains['operations']                                 # TODO: needs to change
        if operations is None:
            operations = [gl.PlaneGroup.E]  # identity matrix
        return np.einsum('ij,mjk->mik', superlattice, operations)
    
    @property
    def primary_beam_theta(self):
        """
        Polar angle of incidence with respect to the direction perpendicular
        to the surface
        """
        return self.__parameters['beamIncidence'][0]
    
    @property
    def primary_beam_phi(self):
        """
        Azimuthal angle of incidence of the primary beam with respect to the
        x axis. Positive counterclockwise when looking down on the surface.
        """
        return self.__parameters['beamIncidence'][1]

    def screen_radius(self, en):
        return gl.screen_radius(en, self.__parameters['screenAperture'])

    def rotate(self, angle, which='surf'):
        """
        This function is a mess and needs to be split into two functions.
        It currently does two different things: if which='bulk', it returns a
        copy of the bulk lattice. If which='surf' it rotates all the LEED
        subpatterns. It would probably be nicer to also have the bulk pattern
        be a special subpattern.
        """
        which = which[0:4]
        if which not in ['surf', 'bulk']:
            raise ValueError("LEEDPattern: Only 'surf', 'surface', or "
                             "'bulk' are acceptable values for rotate()")
        if hasattr(angle, '__len__'):
            raise ValueError('Angle input is not a scalar')

        if which == 'surf':
            angle = np.radians(angle)
            rot = np.array([[np.cos(angle), np.sin(angle)],
                            [-np.sin(angle), np.cos(angle)]])
            for pat in [*self.firstLEED, *self.domsLEED]:
                pat.transformCoords(rot)
            return None
        return self.reciprocal_lattices['bulk'].get_rotated_lattice(angle)


class LEEDStructuralDomains:
    pass

# TODO: LEEDsubpattern
